static char rcsid[] = "@(#)$Id: elm.c,v 1.74 2004/07/27 20:25:02 hurtta Exp $";

/*****************************************************************************
 *  The Elm (ME+) Mail System  -  $Revision: 1.74 $   $State: Exp $
 *
 *  Modified by: Kari Hurtta <hurtta+elm@posti.FMI.FI> 
 *                           (was hurtta+elm@ozone.FMI.FI)
 *****************************************************************************
 *  The Elm Mail System 
 *
 * This file and all associated files and documentation:
 *			Copyright (c) 1988-1992 USENET Community Trust
 *			Copyright (c) 1986,1987 Dave Taylor
 *****************************************************************************/

/* Main program of the ELM mail system! 
*/

#include "patchlevel.h"
#include "elm.h"
#include "s_elm.h"

#ifdef BSD_TYPE
#  include <sys/timeb.h>
#endif

#include "mime.h"
#include "me.h"

DEBUG_VAR(Debug,__FILE__,"ELM");


static unsigned char * s2us P_((char *str));
static unsigned char * s2us(str)
     char *str;
{
    return (unsigned char *)str;
}

static void do_check_only P_((char *argv[],
			      struct AliasView *aview)); 
static void debug_screen P_((struct MailboxView *mailbox,
			     int header_page));
static void debug_message P_((struct header_rec *hdr, int current,
			      struct folder_view *index));

static void elm_panic_prepare P_((CONST int interrupt, 
				  CONST char * title,
				  CONST char * ms));
static void elm_panic_prepare(interrupt,title,ms)
     CONST int interrupt;
     CONST char * title;
     CONST char * ms;
{
    int do_cursor = RawState();

    if (do_cursor && !interrupt) {
	lib_error(FRM("%s: %s"),title,ms);
	sleep(1);
    }

    if (do_cursor) {
	/* MoveCursor() is not safe on interrupt */

	if (interrupt) {
	    ClearScreen();

	    Raw(OFF|RAW_FROM_SIGNAL);
	} else {
	    int LINES, COLUMNS;

	    menu_get_sizes(default_context, &LINES, &COLUMNS);   

	    MoveCursor(LINES-1,0);
	    Raw(OFF);
	}
    }
}

static void elm_panic_exit P_((CONST int interrupt));
static void elm_panic_exit(interrupt)
     CONST int interrupt;
{
    emergency_exit(interrupt);
}

void forget_passphrase() 
{
#ifdef USE_PGP
    if (pgp_void_passphrase())
	lib_error(CATGETS(elm_msg_cat, ElmSet, 
			  ElmPGPPassphraseForgotten,
			  "PGP passphrase forgotten!"));
#endif
#ifdef REMOTE_MBX
    if(close_cached_connections())
	lib_error(CATGETS(elm_msg_cat, ElmSet, 
			  ElmCachedConenctionsClosed,
			  "Cached connections closed!"));
#endif

    error_sleep(1);
}

#include <errno.h>
#ifndef ANSI_C
extern int errno;
#endif         

int mailbox_command(mailbox, ch,  nufoot, cmdX, aview, header_page,
		    page)
     struct MailboxView *mailbox;
     int ch;   
     int *nufoot;
     showmsg_cmd  cmdX;
     struct AliasView *aview;
     int *header_page;
     struct menu_context  *page;
{
    int redraw = 0;
    struct menu_common MENU;
    int LINES, COLUMNS;

    menu_get_sizes(page, &LINES, &COLUMNS);   
    

    set_mcommon_from_mbxview(&MENU,mailbox);

    switch (ch) {

#ifdef ALLOW_SUBSHELL
    case '!'    :  Writechar('!'); 
	subshell(mailbox, page);		
	redraw += menu_need_redraw(page);

	break;
#endif

    case '<'    :  {
	/* scan current message for calendar information */
#ifdef ENABLE_CALENDAR
	int current = get_current(mailbox);

	Write_to_screen(CATGETS(elm_msg_cat, 
				ElmSet, 
				ElmScanForCalendar,
				"Scan message for calendar entries..."));
		
	if  (current < 1) {
	    lib_error(CATGETS(elm_msg_cat, ElmSet, 
			      ElmNoMailToScan,
			      "No mail to scan!"));
	    
	} else if (give_message_data(mailbox,current-1,
				     NULL,NULL,NULL,NO_mime_parse)) {
	    /* FIX: mime sctructure is not 
	       currently used */
	    scan_calendar(mailbox);
	} else {
	    DPRINT(Debug,3,(&Debug, 
			    "give_message_data [%d] fails",current-1));
	}
#else
	lib_error(CATGETS(elm_msg_cat, ElmSet, 
			  ElmSorryNoCalendar,
			  "Sorry. Calendar function disabled."));
	
#endif
    }
	break;
	
    case 'a'    :  
	alias(mailbox, aview);
	redraw++;
	/* define_softkeys(MAIN); */	
	break;

    case 'b'    :  {
	int current = get_current(mailbox);
	struct header_rec *hdr;
	FILE *F;
	
	Write_to_screen(CATGETS(elm_msg_cat, 
				ElmSet, ElmBounceMessage,
				"Bounce message"));
	FlushBuffer();
	if (current < 1) {
	    lib_error(CATGETS(elm_msg_cat, ElmSet, 
			      ElmNoMailToBounce,
			      "No mail to bounce!"));
	    
	} else if (give_message_data(mailbox,current-1,
				     &hdr,&F,NULL,
				     NO_mime_parse)) {
	    (*nufoot) = remail(hdr,F, aview, page);
	    
	    redraw += menu_need_redraw(page);
	} else {
	    DPRINT(Debug,3,(&Debug, 
			    "give_message_data [%d] fails",current-1));
	}
    }
	break;

#ifdef USE_PGP
    case ctrl('E'): {
	int current = get_current(mailbox);
	struct header_rec *hdr;
	FILE *F;

	Write_to_screen(CATGETS(elm_msg_cat, ElmSet, ElmExtractPGPKey,
				"Extract PGP public key"));
	
	if (current < 1) 
	    lib_error(CATGETS(elm_msg_cat, ElmSet, ElmNoMailInFolder,
			      "There is no mail in this folder!"));

	else if (give_message_data(mailbox,current-1,
				     &hdr,&F,NULL,NO_mime_parse)) {
		  
	    if (hdr->pgp & PGP_PUBLIC_KEY) {
		/* FIX: mime structure is not currently used */
		pgp_extract_public_key(hdr,F,page);
		redraw += menu_need_redraw(page);
	    } else
		lib_error(CATGETS(elm_msg_cat, ElmSet, 
				  ElmNoPgpPublicKeys,
				  "This message does not contain PGP public keys!"));
	} else {
	    DPRINT(Debug,3,(&Debug, 
			    "give_message_data [%d] fails",current-1));

	}
    }
	break;
	
#endif

    case ctrl('F'):
	Write_to_screen(CATGETS(elm_msg_cat, ElmSet, 
				ElmPassphraseForget,
				"Forget passphrase"));
	forget_passphrase();
	break;

    case 'f'    : {
	int current = get_current(mailbox);
	struct header_rec *hdr;
	FILE *F;
	
	Write_to_screen(CATGETS(elm_msg_cat, ElmSet, 
				ElmForward,
				"Forward"));
	/* define_softkeys(YESNO); */
	FlushBuffer();
	
	if (current < 1)
	    lib_error(CATGETS(elm_msg_cat, ElmSet, 
			      ElmNoMailToForward,
			      "No mail to forward!"));
		  
	else  {
	    forward(current-1,mailbox, aview, page);
	    	   
	    if (menu_need_redraw(page))
		redraw++;
	    else {
		(*nufoot)++;
		show_msg_status(current-1,&MENU, *header_page);
	    }
	} 
	/* define_softkeys(MAIN); */
    }
	break;

    case 'g'    :  {
	int current = get_current(mailbox);
	struct header_rec *hdr;
	FILE *F;


	Write_to_screen(CATGETS(elm_msg_cat, ElmSet, 
				ElmGroupReply,
				"Group reply"));
	FlushBuffer();
	
	if (current < 1)
	    lib_error(CATGETS(elm_msg_cat, ElmSet, 
			      ElmNoMailToReply,
			      "No mail to reply to!")); 
	else if (give_message_data(mailbox,current-1,
				   &hdr,NULL,NULL,
				   mime_parse_routine)) {
	    
	    if (hdr->status & FORM_LETTER) {
		lib_error(CATGETS(elm_msg_cat, ElmSet, 
				  ElmCantGroupReplyForm,
				  "Can't group reply to a Form!!"));
		
	    } else {
		/* define_softkeys(YESNO); */
		
		reply_to_everyone(current-1, mailbox, 
				  aview,page);	
		redraw += menu_need_redraw(page);
		/* define_softkeys(MAIN); */

		if (!redraw)
		    show_msg_status(current-1,&MENU, *header_page);
	    }
	} else {
	    DPRINT(Debug,3,(&Debug, 
			    "give_message_data [%d] fails",current-1));

	}
	FlushBuffer();
    }
	break;
	
    case 'h'    :  {
	int current = get_current(mailbox);
		
	if (elm_filter)
	    Write_to_screen(CATGETS(elm_msg_cat, 
				    ElmSet, 
				    ElmMessageWithHeaders,
				    "Message with headers..."));
	else
	    Write_to_screen(CATGETS(elm_msg_cat, 
				    ElmSet, 
				    ElmDisplayMessage,
				    "Display message"));
	FlushBuffer();

	if (current < 1)
	    lib_error(CATGETS(elm_msg_cat, ElmSet, 
			      ElmNoMailToRead,
			      "No mail to read!"));
	else {		
	    struct header_rec *hdr;
	    FILE *F;
	    int j = elm_filter;
	    int i = 0;

	    elm_filter = FALSE;
	    

	    if (give_message_data(mailbox,current-1,
				    &hdr,&F,NULL,
				    mime_parse_routine))
		i = show_msg(hdr,F,current,
			     get_message_count(mailbox),
			     page);
	    else {
		DPRINT(Debug,3,(&Debug, 
				"give_message_data [%d] fails",current-1));
	    }

	    while (i > 0) {
		i = process_showmsg_cmd(i, mailbox, cmdX, aview, header_page,
					page);
		
		switch (i) {
		case EOF:
		    return EOF;/* Read failed, control tty died? */
		    
		case 'X':      /* Quick Exit */
		case 'x':      /* Exit       */
		    
		    return i;   /* "call" parent */
		case 0:
		    break;

		default:
		    DPRINT(Debug,3,(&Debug, 
				    "Looping showmsg command %d (%c)\n",
				    i,i));
		}
	    }


	    current = get_current(mailbox);

	    elm_filter = j;
	    redraw++;
	    (void)get_page(current, &MENU, header_page);
	} 
    }
	break;
	
#ifdef USE_PGP
    case ctrl('K'):
	Write_to_screen(CATGETS(elm_msg_cat, ElmSet,
				ElmMailPGPPublicKey,
				"Mail PGP public key"));
	FlushBuffer();
	
	pgp_mail_public_key(mailbox, aview, page);
	redraw += menu_need_redraw(page);

	break;
#endif
	
    case 'm'    :   Write_to_screen(CATGETS(elm_msg_cat, ElmSet,
					    ElmMail,
					    "Mail"));
	FlushBuffer();
	send_msg_l(-1, NULL, NULL, NULL,
				MAIL_EDIT_MSG,allow_forms,
				mailbox, aview,page);

	redraw += menu_need_redraw(page);

	break;
			   
    case ' '    : 
    case ctrl('J'):
    case ctrl('M'): {
	int current = get_current(mailbox);
		
	Write_to_screen(CATGETS(elm_msg_cat, ElmSet,
				ElmDisplayMessage,
				"Display message"));  
	FlushBuffer();
	
	if (current < 1)
	    lib_error(CATGETS(elm_msg_cat, ElmSet, 
			      ElmNoMailToRead,
			      "No mail to read!"));
	else {
	    struct header_rec *hdr;
	    FILE *F;
	    int i = 0;	    

	    /* define_softkeys(READ); */
	    

	    if (give_message_data(mailbox,current-1,
				    &hdr,&F,NULL,
				    mime_parse_routine))
		i = show_msg(hdr,F,current,
			     get_message_count(mailbox),
			     page);
	    else {
		DPRINT(Debug,3,(&Debug, 
				"give_message_data [%d] fails",current-1));
	    }

	    while (i > 0) {
		i = process_showmsg_cmd(i, mailbox, cmdX, aview, header_page,
					page);

		switch (i) {
		case EOF:
		    return EOF;/* Read failed, control tty died? */
		    
		case 'X':      /* Quick Exit */
		case 'x':      /* Exit       */
		    
		    return i;   /* "call" parent */
		case 0:
		    break;
		    
		default:
		    DPRINT(Debug,3,(&Debug, 
				    "Looping showmsg command %d (%c)\n",
				    i,i));		    
		}
	    }

	    current = get_current(mailbox);

	    redraw++;
	    (void)get_page(current, &MENU, header_page);
	}
    }
	break;
	
    case 'n'    : {
	int current = get_current(mailbox);
	
	Write_to_screen(CATGETS(elm_msg_cat, ElmSet,
				ElmNextMessage,
				"Next Message"));
	FlushBuffer();
	/* define_softkeys(READ); */
	
	if (current < 1)
	    lib_error(CATGETS(elm_msg_cat, ElmSet, 
			      ElmNoMailToRead,
			      "No mail to read!"));
	else {
	    struct header_rec *hdr;
	    FILE *F;
	    int i = 0;	    
	    int mc;
	    
	    /* define_softkeys(READ); */
	    if (give_message_data(mailbox,current-1,
				    &hdr,&F,NULL,
				    mime_parse_routine))
		i = show_msg(hdr, F, current,
			     get_message_count(mailbox),
			     page);
	    else {
		DPRINT(Debug,3,(&Debug, 
				"give_message_data [%d] fails",current-1));
	    }

	    while (i >0) {
		i = process_showmsg_cmd(i, mailbox, cmdX, aview, header_page,
					page);
	    
		switch (i) {
		case EOF:
		    return EOF;/* Read failed, control tty died? */
		    
		case 'X':      /* Quick Exit */
		case 'x':      /* Exit       */
		    
		    return i;   /* "call" parent */

		case 0:
		    break;
		    
		default:
		    DPRINT(Debug,3,(&Debug, 
				    "Looping showmsg command %d (%c)\n",
				    i,i));
		}
	    }

	    current = get_current(mailbox);
	    
	    redraw++;
	    if (++current > (mc = get_message_count(mailbox)))
		current = mc;
	    set_current(mailbox,current);

	    (void)get_page(current, &MENU, header_page);
	} 
    }		
	break;

    case ctrl('N'):   {
	int current = get_current(mailbox);
	
	Write_to_screen(CATGETS(elm_msg_cat, ElmSet,
				ElmNextUMessage,
				"Next Unread Message"));
	FlushBuffer();
	
	
	if (current < 1)
	    lib_error(CATGETS(elm_msg_cat, ElmSet, 
			      ElmNoMailToRead,
			      "No mail to read!"));
	else {
	    struct header_rec * hdr = give_header(mailbox,
						  current-1);
	    int mc;
	    int i = 0;	    

	    if(ison(hdr->status, UNREAD)) {
		struct header_rec *hdr;
		FILE *F;
		

		if (give_message_data(mailbox,current-1,
					&hdr,&F,NULL,
					mime_parse_routine))
		    i = show_msg(hdr,F,current,
				 get_message_count(mailbox),
				 page);
		else {
		    DPRINT(Debug,3,(&Debug, 
				    "give_message_data [%d] fails",current-1));
		}

	    } else 
		i = ctrl('N');

	    while (i >0) {
		i = process_showmsg_cmd(i, mailbox, cmdX, aview, header_page,
					page);
		
		switch (i) {
		case EOF:
		    return EOF;/* Read failed, control tty died? */
		    
		case 'X':      /* Quick Exit */
		case 'x':      /* Exit       */
		    
		    return i;   /* "call" parent */

		case 0:
		    break;
		    
		default:
		    DPRINT(Debug,3,(&Debug, 
				    "Looping showmsg command %d (%c)\n",
				    i,i));		    
		}
	    }

	    current = get_current(mailbox);

	    redraw++;
	    if (++current > (mc = get_message_count(mailbox)))
		current = mc;
	    set_current(mailbox,current);
	    
	    (void)get_page(current, &MENU, header_page);
	}
    }
	break;
	
    case 'O' : {
	int current = get_current(mailbox);
	
	Write_to_screen(CATGETS(elm_msg_cat, ElmSet,	  
				ElmOverrideCharset,
				"Override charset"));
	FlushBuffer();

	if  (current < 1) 
	    lib_error(CATGETS(elm_msg_cat, ElmSet, 
			      ElmNoMailToOverride,
			      "No mail to override!"));
	else
	    OverrideCharset(mailbox, *header_page, page);

    }
	break;
    case 'P'    : {
	int current = get_current(mailbox);
	
	Write_to_screen(CATGETS(elm_msg_cat, ElmSet,	  
				ElmPrintTextMail,
				"Print text of mail"));
	FlushBuffer();
	
	if (current < 1) {
	    lib_error(CATGETS(elm_msg_cat, ElmSet, 
			      ElmNoMailToPrint,
			      "No mail to print!"));
	    
	} else if (give_message_data(mailbox,current-1,
				     NULL,NULL,NULL,
				     mime_parse_routine)) {
	    print_text(TRUE, mailbox, *header_page, page);
	    if (menu_need_redraw(page))
		redraw++;
	    else
		show_msg_status(current-1,&MENU, *header_page);
	} else {
	    DPRINT(Debug,3,(&Debug, 
			    "give_message_data [%d] fails",current-1));
	}
    }
	break;
	
    case 'r'    : {
	int current = get_current(mailbox);
	struct header_rec *hdr;
	FILE *F;
	
	Write_to_screen(CATGETS(elm_msg_cat, ElmSet,
				ElmReplyToMessage,
				"Reply to message"));
	FlushBuffer();
	
	if (current < 1)
	    lib_error(CATGETS(elm_msg_cat, ElmSet, 
			      ElmNoMailToReplyTo,
			      "No mail to reply to!")); 
	else {
	    
	    reply(current-1,mailbox, aview, page);	
	    redraw += menu_need_redraw(page);

	    if (!redraw)
		show_msg_status(current-1,&MENU, *header_page);
	} 
	/* softkeys_on(); */
    }
	break;
	
    case '>'    : /** backwards compatibility **/
	
    case 'C'	:
    case 's'    :  {
	int current = get_current(mailbox);
	
	if (ch != 'C')
	    Write_to_screen(CATGETS(elm_msg_cat, ElmSet,
				    ElmSaveToFolder,
				    "Save to folder"));
	else
	    Write_to_screen(CATGETS(elm_msg_cat, ElmSet,
				    ElmCopyToFolder,
				    "Copy to folder"));
	FlushBuffer();
	
	if  (current < 1) {
	    if (ch != 'C')
		lib_error(CATGETS(elm_msg_cat, ElmSet, 
				  ElmNoMailToSave,
				  "No mail to save!"));
	    else
		lib_error(CATGETS(elm_msg_cat, ElmSet, 
				  ElmNoMailToCopy,
				  "No mail to copy!"));
	    
	} else if (give_message_data(mailbox,current-1,
				     NULL,NULL,NULL,NO_mime_parse)) {
	    
	    if (save(FALSE, (ch != 'C'), 0,mailbox, aview, 
		     *header_page, page)
		&& resolve_mode && ch != 'C') {
		int i;

		if((i=next_message(current-1, TRUE, &MENU)) != -1) {
		    
		    int current        = get_current(mailbox);
		    current = i+1;
		    set_current(mailbox,current);
		    
		    nucurr = get_page(current,&MENU, header_page);
		    
		}
	    }

	    redraw += menu_need_redraw(page);
	} else {
	    DPRINT(Debug,3,(&Debug, 
			    "give_message_data [%d] fails",current-1));
	}
	ClearLine(LINES-3);		
    }
	break;

    case 'S'     :   { 
	int current = get_current(mailbox);
	
	Write_to_screen(CATGETS(elm_msg_cat, ElmSet,
				ElmSaveTextToFileorFolder,
				"Save (copy) text to file or folder"));
	FlushBuffer();
	
	if  (current < 1) {
	    lib_error(CATGETS(elm_msg_cat, ElmSet, 
			      ElmNoMailToCopy,
			      "No mail to copy!"));
	    
	} else if (give_message_data(mailbox,current-1,
				     NULL,NULL,NULL,
				     mime_parse_routine)) {
	    save(FALSE, 0, 1, mailbox, aview, *header_page,
		 page);

	    redraw += menu_need_redraw(page);

	    if (!redraw)
		show_msg_status(current-1,&MENU, *header_page);
	} else {
	    DPRINT(Debug,3,(&Debug, 
			    "give_message_data [%d] fails",current-1));
	}
	
	ClearLine(LINES-3);		
    }
	break;	                   

    case 'U'    :  {
	if (get_message_count(mailbox) < 1) {
	    lib_error(CATGETS(elm_msg_cat, ElmSet,
			      ElmNoItemToMarkUnread,
			      "No %S to mark as unread!"),
		      mcommon_give_item(&MENU,m_items));
	    
	}
	else
	    unread_msg(TRUE, mailbox, *header_page);
    }
	break;
	
    case 'v'    :   {	      
	struct header_rec *hdr;
	FILE *F;
	int current = get_current(mailbox);
	
	if (current < 1)
	    lib_error(CATGETS(elm_msg_cat, ElmSet, 
			      ElmNoMessages,
			      "There are no messages!"));
	else {
	    Write_to_screen(CATGETS(elm_msg_cat, ElmSet, 
				    ElmViewParts,
				    "View parts of message"));
	    
	    if (give_message_data(mailbox,current-1,
				  &hdr,&F,NULL,mime_parse_routine)) {
		if ((hdr->status & MIME_MESSAGE)) {
		    
		    mime_warnings(hdr);
		    
		    attach_menu(F,& (hdr->mime_rec),
				NULL, 
				hdr->header_charset, NULL,
				hdr,
				aview
				);
		    redraw++;
		} else
		    lib_error(CATGETS(elm_msg_cat, ElmSet, 
				      ElmNoMIMEMessage,
				      "This is not a MIME message!"));
		show_msg_status(current-1,&MENU,*header_page);
	    } else {
		DPRINT(Debug,3,(&Debug, 
				"give_message_data [%d] fails",current-1));
	    }
	}
    }	   	
	break;

    case 'V': {
	int current = get_current(mailbox);

	Write_to_screen(CATGETS(elm_msg_cat, ElmSet,
				ElmViewDigest,
				"View digest as mailbox"));

	FlushBuffer();
	
	if (current < 1)
	    lib_error(CATGETS(elm_msg_cat, ElmSet, 
			      ElmNoMailToView,
			      "No mail to view!")); 
	else if (give_message_data(mailbox,current-1,
				   NULL,NULL,NULL,mime_parse_routine)) {
	    
	    ViewDigest(mailbox,aview);

	    current = get_current(mailbox);
	    (void)get_page(current, &MENU, header_page);
	    redraw++;

	} else {
	    DPRINT(Debug,3,(&Debug, 
			    "give_message_data [%d] fails",current-1));
	    show_msg_status(current-1,&MENU,*header_page);
	}
    }
	break;

    case EOF : 
	return EOF;/* Read failed, control tty died? */
     
    case '@'    : 
	debug_screen(mailbox, *header_page);  
	redraw++;	
	break;

    case '#'    : {
	int current = get_current(mailbox);
	Write_to_screen(CATGETS(elm_msg_cat, ElmSet, 
				ElmCheckMessage,
				"Check message"));
	FlushBuffer();
	
	if (current < 1)
	    lib_error(CATGETS(elm_msg_cat, ElmSet, 
			      ElmNoMailToCheck,
			      "No mail to check."));
	else {
	    struct header_rec * hdr = give_header(mailbox,
						  current-1);
	    if (hdr) {
		struct folder_view index;
		
		give_index_number(mailbox,current-1,
				  &index);
		
		debug_message(hdr, current, &index); 
		redraw++;	
	    }
	} 
    }
	break;

    default: 
	/* Maybe motion command */

	ch = motion(ch,&MENU,nufoot, header_page,
		    page);
    
	/* motion() may already set, menu_trigger_redraw(page),
           we do not examine it */
	if (redraw)
	    menu_trigger_redraw(page);

	return ch;
		
    }

    if (redraw)
	menu_trigger_redraw(page);
    return 0;   /* Command OK */
}

void new_mail_check (mailbox, header_page, page)
     struct MailboxView *mailbox;
     int *header_page;
     struct menu_context  *page;
{
    int idx, mbxcount;
    struct menu_common MENU;
    int   LINES, COLUMNS;	
    int redraw = 0;

    menu_get_sizes(page,&LINES, &COLUMNS);

    set_mcommon_from_mbxview(&MENU,mailbox);
    
    mbxcount = get_storage_count(mailbox);

    for (idx = 0 ; idx < mbxcount; idx++) {
	int bytes;

	struct current_storage * storage = get_storage(mailbox, idx);
	    
	if (!storage ||
	    !storage->current_folder)
	    continue;

	flush_folder(storage->current_folder);

	if (new_mail_on_folder(storage->current_folder,
			       &bytes)) {

	    int current        = get_current(mailbox);
	    int was_on_page    = on_page(current, &MENU, *header_page);
	    int pageon         = *header_page;
	    int last_in_folder = storage->message_count;
	    int selected       = get_selected(mailbox);
	    int ok;

	    DPRINT(Debug,2,(&Debug, 
			    "Just received %d bytes more mail (elm), storage idx %d\n", 
			    bytes, idx));
	    
	    lib_transient(CATGETS(elm_msg_cat, ElmSet, ElmNewMailHangOn,
				  "New mail has arrived! Hang on..."));
	    
	    ok = read_new_mails(storage, page);

	    if (ok)
		clear_error();
	    *header_page = pageon;
		  
	    if (selected) {               
		/* update count of selected messages */
		selected += storage->message_count - last_in_folder;
		set_selected(mailbox,selected);
	    }		  

	    if (on_page(current, &MENU, *header_page) || 
		was_on_page || !ok) /* do we REALLY have to rewrite? */
		redraw++;
		  
	    ClearLine(LINES-2);	     /* remove reading message... */
	    
	    if (ok) {
		if ((storage->message_count - last_in_folder) == 1)
		    lib_error(CATGETS(elm_msg_cat, ElmSet, ElmNewMessageRecv,
				      "1 new message received."));
		else
		    lib_error(CATGETS(elm_msg_cat, ElmSet, 
				      ElmNewMessageRecvPlural,
				      "%d new messages received."), 
			      storage->message_count - last_in_folder);
	    } else
		show_last_error();
	    
	} else if (bytes < 0) {
	    lib_error(CATGETS(elm_msg_cat, ElmSet, 
			      ElmALERTMailfileShrinken,
			      "ALERT: Mailfile shrinken %d bytes!"),
		      -bytes);
	    DPRINT(Debug,1,(&Debug, 
			    "ALERT: Mailfile shrinken %d bytes!",
			    -bytes));
	    sleep_message();
	}
    }
    
    if (redraw)
	menu_trigger_redraw(page);

}



int main(argc, argv)
     int argc;
     char *argv[];
{
    int  ch;
    char **req_mfile_vector = NULL;    /* Malloced */
    char **to_whom = NULL;
    int auto_assemble = 1;
    int  i,j;      		/** Random counting variables (etc)          **/

    int redraw = 0,	 		/* do we need to rewrite the entire screen? */
	nufoot = 0;	 		/* clear lines 16 thru bottom and new menu  */
    int header_page = 0;     	/* current header page     */

    struct menu_context  *page = NULL;
    struct MailboxView   * cur_mailbox = NULL;
    struct AliasView     * cur_aliaslist = NULL;

    set_panic_prepare(elm_panic_prepare);
    set_panic_exit(elm_panic_exit);

#if DEBUG
    init_debugfile("ELM");

    /* HACK .... 
       uSER may want enable debug output for locale_init()
       
       This gives wrong result if some options value start
       with "-d"
    */

    for (i = 1; i < argc; i++) {
	if ('-' == argv[i][0] &&
	    'd' == argv[i][1] &&
	    argv[i][2]) {
	    set_debugging(& (argv[i][2]));
	}
    }
#endif	

    locale_init();
	
    req_mfile_vector = parse_arguments(argc, argv, &to_whom);
    initialize(req_mfile_vector, &cur_mailbox, &cur_aliaslist, 
	       &header_page, &page);
    if (req_mfile_vector) {
	int i;

	for (i = 0; req_mfile_vector[i]; i++) {
	    free(req_mfile_vector[i]);
	    req_mfile_vector[i] = NULL;
	}

	free(req_mfile_vector);
	req_mfile_vector = NULL;
    }



    if (mail_only) {
	DPRINT(Debug,2,(&Debug, 
			"Mail-only: mailing to\n"));
	
	if(!batch_only) {
	    print_format_center(1,
				CATGETS(elm_msg_cat, ElmSet, ElmSendOnlyMode,
					"Send only mode [ELM %s]"), 
				version_buff);
	}
	send_msg_argv(to_whom, batch_subject, MAIL_EDIT_MSG,
		      (batch_only ? NO : allow_forms), cur_mailbox,
		      cur_aliaslist, page);

	leave(0);
    } else if (check_only) {
	do_check_only(to_whom, cur_aliaslist);
	leave(0);
    }

    showscreen(cur_mailbox,
	       show_menu, &header_page, page);

    while (1) {

	struct menu_common MENU;
    	int   LINES, COLUMNS;	

	menu_get_sizes(page,&LINES, &COLUMNS);

	set_mcommon_from_mbxview(&MENU,cur_mailbox);

	if (menu_resized(page)) {
	    
	    menu_get_sizes(page,&LINES, &COLUMNS);

#undef max
#define max(a,b)	       ((a) < (b) ? (b) : (a))
	    if (mini_menu)
		headers_per_page = max (LINES - 14, 1);
	    else
		headers_per_page = max (LINES -	 9, 1);	  /* 5 more headers! */
#undef max
		redraw++;
	}	
	else 
	    redraw += menu_need_redraw(page);

	nufoot = 0;
	nucurr = 0;

#ifdef BACKGROUD_PROCESSES      
	if (handle_sigchld)
	    sigchld_handler();
#endif

	new_mail_check(cur_mailbox,&header_page,
		       page);
	redraw += menu_need_redraw(page);


	/* ??? FIX    Is this correct ?? */
	if (update_view(cur_mailbox)) {
	    resort_mailbox(cur_mailbox,1,&header_page);

	    redraw++;
	}

	if (redraw) {
	    DPRINT(Debug,7, (&Debug, 
			     "main: pending redraw\n"));

	    redraw = 0;
	    showscreen(cur_mailbox,
		       show_menu, &header_page, page);
	    show_last_error();
	}
	
	
	{
	    int lin,col;
	    
	    prompt_s(mcommon_give_item(&MENU, m_Prompt));
	    GetXYLocation(&lin,&col);

	    CleartoEOS();
	    show_last_error();
	    MoveCursor(lin,col);

	    /* fragment_handling: 0 == none,
	       1 == manual,
	       2 == auto  */


	    if (auto_assemble && 2 == fragment_handling &&
		have_partial(cur_mailbox)) {
		ch = 'A';
		auto_assemble = 0;
		
		DPRINT(Debug,4,(&Debug,"Automatic assemble ... \n"));
	    } else
		ch = GetPrompt(page);
	
	    /* ? ? ?    resize again? */ 

	    CleartoEOS();

	    if (isascii(ch) && isprint(ch)) {
		DPRINT(Debug,4,(&Debug, 
				"\nCommand: %c [%d]\n\n", ch, ch));
	    } else {
		DPRINT(Debug,4,(&Debug, 
				"\nCommand: %d\n\n", ch));
	    }
	    
	    set_error("");	/* clear error buffer */
	    
	    MoveCursor(lin,col);
	}

	  ch = mailbox_command(cur_mailbox,ch, &nufoot,
			       showmsg_main_cmd,
			       cur_aliaslist,
			       &header_page,page);

	  redraw += menu_need_redraw(page);

	  switch (ch) {
	  case 0:  /* OK */
	      break;
	      
	  case HELP_MARK:
	  case '?' 	:  
	      help(FALSE, page);

	      if (menu_need_redraw(page))
		  redraw++;
	      else
		  nufoot++;
	      break;
	      
	  case '$'    :  {
	      int current;

	      Write_to_screen(CATGETS(elm_msg_cat, ElmSet, 
				      ElmResyncFolder,
				      "Resynchronize folder"));
	      FlushBuffer();

	      resync(& cur_mailbox, &header_page, page);
	      redraw += menu_need_redraw(page);

	      /* FIX MENU pointer ... */
	      set_mcommon_from_mbxview(&MENU,cur_mailbox);

	      current = get_current(cur_mailbox);
	      nucurr = get_page(current, &MENU, &header_page);
	  }
	      break;

	  case '|'    :   {
	      int current = get_current(cur_mailbox);
	      
	      Writechar('|');
	      
	      if (current < 1) {
		  lib_error(CATGETS(elm_msg_cat, ElmSet, 
				    ElmNoMailToPipe,
				    "No mail to pipe!"));
		  
	      } else if (give_message_data(cur_mailbox,current-1,
					   NULL,NULL,NULL,NO_mime_parse)) {
		  
		  /* softkeys_off(); */
		  
		  do_pipe(cur_mailbox, page);		
		  redraw += menu_need_redraw(page);

		  /* softkeys_on(); */
	      } else {
		  DPRINT(Debug,3,(&Debug, 
				  "give_message_data [%d] fails",current-1));
	      }
	    }
	      break;

	  case 'A':    {
	      int current = get_current(cur_mailbox);
	      
	      Write_to_screen(CATGETS(elm_msg_cat, 
				      ElmSet, ElmAssemble,
				      "Assemble message fragments"));
	      FlushBuffer();
	      sleep_message();

	      ViewPartial(cur_mailbox, cur_aliaslist);

	      current = get_current(cur_mailbox);
	      (void)get_page(current, &MENU, &header_page);
	      redraw++;

	  }
	      break;

	  case 'c'    :  Write_to_screen(CATGETS(elm_msg_cat, 
						 ElmSet, ElmChangeFolder,
						 "Change folder"));
	      /* define_softkeys(CHANGE); */
	      change_file(& cur_mailbox, cur_aliaslist,
				    &header_page, page);
	      redraw += menu_need_redraw(page);


	      /* FIX MENU pointer ... */
	      set_mcommon_from_mbxview(&MENU,cur_mailbox);

	      /* define_softkeys(MAIN); */
	      break;

#ifdef ALLOW_MAILBOX_EDITING
	  case 'e'    :  {
	      int current = get_current(cur_mailbox);
	      
	      Write_to_screen(CATGETS(elm_msg_cat, 
				      ElmSet, ElmEditFolder,
				      "Edit folder"));
	      
	      if (user_level < 2) {
		  lib_error(CATGETS(elm_msg_cat, ElmSet, 
				    ElmNoFolderEditing,
				    "You are not experienced to use folder editing"));
		  
		  break;
	      }
	      
	      if (current < 1) 
		  lib_error(CATGETS(elm_msg_cat, ElmSet, 
				    ElmFolderIsEmpty,
				    "Folder is empty!"));
	      else
		  edit_mailbox(&cur_mailbox, &header_page,
			       page);
	      
	  }
	      break;
#else
	  case 'e'    : lib_error(CATGETS(elm_msg_cat, ElmSet, 
					  ElmNoFolderEdit,
					  "Folder editing isn't configured in this version of ELM."));
	      
	      break;
#endif

	  case 'o'    :  {
	      int current = get_current(cur_mailbox);
	      
	      Write_to_screen(CATGETS(elm_msg_cat, ElmSet,
				      ElmOptions,
				      "Options"));
	      if((i=options(cur_mailbox, cur_aliaslist)) > 0) {
		  current = get_current(cur_mailbox);
		  get_page(current, &MENU, &header_page);
	      } else if(i < 0)
		  leave(0);
	      redraw++;	/* always fix da screen... */
	  }
	      break;

	  case 'p'    :  {
	      int current = get_current(cur_mailbox);
	      
	      Write_to_screen(CATGETS(elm_msg_cat, ElmSet,
				      ElmPrintMail,
				      "Print mail"));
	      FlushBuffer();

#if DISABLE_PRINT
	      lib_error(FRM("p)rint currently disabled, try P)rint text instead"));
#else
	      if (current < 1) {
		  lib_error(CATGETS(elm_msg_cat, ElmSet, 
				    ElmNoMailToPrint,
				    "No mail to print!"));
		
	      } else if (give_message_data(cur_mailbox,current-1,
					   NULL,NULL,NULL,NO_mime_parse)) {
		  
		  /* FIX: mime structure is not currenty
		     used
		  */
		  print_msg(TRUE,cur_mailbox, page);
		   
		  redraw += menu_need_redraw(page);

	      } else {
		  DPRINT(Debug,3,(&Debug, 
				  "give_message_data [%d] fails",current-1));
	      }
#endif
	  }
	      break;


	    case 'q'    :  {
		int idx, mbxcount, found = 0;

		Write_to_screen(CATGETS(elm_msg_cat, ElmSet,
					ElmQuit,
					"Quit"));
		FlushBuffer();

		mbxcount = get_storage_count(cur_mailbox);
		
		for (idx = 0 ; idx < mbxcount; idx++) {
		    int bytes;

		    struct current_storage * storage = 
			get_storage(cur_mailbox,idx);
	    
		    if (!storage->current_folder)
			continue;


		    if (new_mail_on_folder(storage->current_folder,&bytes)) {
			lib_error(CATGETS(elm_msg_cat, ElmSet, 
					  ElmNewMailQuitCancelled,
					  "New Mail!  Quit canceled..."));
			found++;
		    }
		}

		if (!found)
		    quit(TRUE, &cur_mailbox,&cur_aliaslist,
			 &header_page, page);		
	    }			   
		break;

	  case 'Q'    :  {
	      int idx, mbxcount, found = 0;

	      Write_to_screen(CATGETS(elm_msg_cat, ElmSet,
				      ElmQuickQuit,
				      "Quick quit"));
	      FlushBuffer();

	      mbxcount = get_storage_count(cur_mailbox);
	      
	      for (idx = 0 ; idx < mbxcount; idx++) {
		  int bytes;

		  struct current_storage * storage = 
		      get_storage(cur_mailbox,idx);
		  
		  if (!storage->current_folder)
		      continue;

		  if (new_mail_on_folder(storage->current_folder,&bytes)) {
		      lib_error(CATGETS(elm_msg_cat, ElmSet, 
					ElmNewMailQuickQuitCancelled,
					"New Mail!  Quick Quit canceled..."));	
		      found++;
		  }
	      }

	      if (!found)
		  quit(FALSE, &cur_mailbox,&cur_aliaslist,
		       &header_page, page);		
	  }
	      break;


		

	      
	  case 'X'    :  Write_to_screen(CATGETS(elm_msg_cat, ElmSet,
						 ElmQuickExit,
						 "Quick Exit"));
	                 FlushBuffer();
	                 leave(0);
			 break;
		
	    case ctrl('Q') :
	  case 'x'    :    Write_to_screen(CATGETS(elm_msg_cat, ElmSet,
						   ElmExit,
						   "Exit"));  
                           FlushBuffer();

			   exit_prog(&cur_mailbox,&cur_aliaslist,
				     &page);
			   break;

	  case EOF :  
	      leave(0);  
	      /* Read failed, control tty died? */
	      break;
	    
	
	  default:
	      lib_error(CATGETS(elm_msg_cat, ElmSet, 
				ElmUnknownCommand,
				"Unknown command. Use '?' for help."));		    
	  }

	  if (redraw)
	      showscreen(cur_mailbox,
			 show_menu, &header_page, page);

          check_range(&MENU);
	    
	  if (nucurr == NEW_PAGE) {
	      DPRINT(Debug,5,(&Debug,  " nucurr == NEW_PAGE\n"));
	      show_headers(&MENU,&header_page, page);
	  } else if (nucurr == SAME_PAGE) {
	      DPRINT(Debug,5,(&Debug,  " nucurr == SAME_PAGE\n"));
	      show_current(&MENU,&header_page, page);
	  } else if (nufoot) {
	    if (mini_menu) {
		MoveCursor(LINES-8, 0);  
		CleartoEOS();
		show_menu(page);
	    } else {
	      MoveCursor(LINES-5, 0);
	      CleartoEOS();
	    }
	    show_last_error();	/* for those operations that have to
				 * clear the footer except for a message.
				 */
	  }

	} /* the BIG while loop! */
}

static void debug_screen(mailbox, header_page)
     struct MailboxView *mailbox;
     int header_page;
{
	/**** spit out all the current variable settings and the table
	      entries for the current 'n' items displayed. ****/

	int i, j;
	int LINES, COLUMNS;
	struct menu_context *page = new_menu_context();

redraw:
	if (menu_resized(page) ||
	    menu_need_redraw(page))
	/* Nothing */;
        
	menu_get_sizes(page, &LINES, &COLUMNS);   

	menu_ClearScreen(page);

	PutLineX(0,0,CATGETS(elm_msg_cat, ElmSet, 
			     ElmDbxCurrMsgNum,
			     "Current message number = %d\t\t%d message(s) total\n"),
		 get_current(mailbox), 
		 get_message_count(mailbox));
	PutLineX(2,0,CATGETS(elm_msg_cat, ElmSet, 
			     ElmDbxHdrPage,
			     "Header_page = %d           \t\t%d possible page(s)\n"),
		 header_page, (int) (get_message_count(mailbox) / 
				     headers_per_page) + 1);

	{ 

	    struct folder_info * f = get_main_folder(mailbox);
	    if (f) {
		PutLineX(4,0,CATGETS(elm_msg_cat, ElmSet, 
				     ElmDbxCurrMailFile,
				     "\nCurrent mail file is %S.\n\r\n"), 
			 f->cur_folder_disp);
	    }
	}

	i = header_page*headers_per_page;	/* starting header */

	if ((j = i + (headers_per_page-1)) >= get_message_count(mailbox)) 
	    j = get_message_count(mailbox)-1;

	Write_to_screen(CATGETS(elm_msg_cat, ElmSet, ElmDebugHeader,
				"Num      From                    Subject                Lines   Offset  Content\n\r\n\r"));
	
	for (;i <= j; i++) {	  
	    struct header_rec * hdr = give_header(mailbox,i);
	    struct string * x    = NULL;
	    struct string * temp = NULL;
	    

	    if (! hdr)
		continue;
	    
	    x = hdr->subject;
	    if (!x) {
		temp = new_string2(ASCII_SET,s2us("<-- NO SUBJECT -->"));
		x = temp;
	    }

	    if (hdr->from && 
		hdr->from->fullname &&
		string_len(hdr->from->fullname))
		Write_to_screen(FRM("%3d %-16.16S %-35.35S %4d %8d %8d\n\r"),
				i+1,
				hdr->from->fullname, 
				x,
				hdr->lines,
				hdr->offset,
				hdr->content_length);
	    else
		Write_to_screen(FRM("%3d %-16.16s %-35.35S %4d %8d %8d\n\r"),
				i+1,
				hdr->from && 
				hdr->from->addr &&
				hdr->from->addr[0] ?
				hdr->from->addr :
				hdr->env_from,
				x,
				hdr->lines,
				hdr->offset,
				hdr->content_length);

	    if (temp)
		free_string(&temp);

	}
	
	PutLineX(LINES-1,0,CATGETS(elm_msg_cat, ElmSet, 
			     ElmDbxPressAnyKey,"Press any key to return."));
	if (REDRAW_MARK == menu_ReadCh(page, REDRAW_MARK))
	  goto redraw;

  erase_menu_context(&page);
  return;

}

static void debug_message(current_header,current,index)
     struct header_rec *current_header;
     int current;
     struct folder_view *index;
{
	/**** Spit out the current message record.  Include EVERYTHING
	      in the record structure. **/
	
	time_t header_time;
	int LINES, COLUMNS;
	struct menu_context *page = new_menu_context();
	
redraw:
	if (menu_resized(page) ||
	    menu_need_redraw(page))
	    /* Nothing */;
        
	menu_get_sizes(page, &LINES, &COLUMNS);   

	menu_ClearScreen(page);

	Write_to_screen(FRM("\t\t\t----- Message %d -----\n\r\n\r\n\r\n\r"), 
			current);

	Write_to_screen(FRM("Lines: %-17dStatus: A  C  D  E  F  N  P  T  U  V  O  R  M  P  U  N\n\r"), 
			current_header->lines);
	Write_to_screen(FRM("Content-Length: %-16dc  o  e  x  o  e  r  a  r  i  l  e  i  r  s  H\n\r"), 
			current_header->content_length);
	Write_to_screen(FRM("Binary: %-24dt  n  l  p  r  w  i  g  g  s  d  p  m  e  u  d\n\r"), 
			current_header->binary);
	Write_to_screen(FRM("'From ' on body: %-15dn  f  d  d  m     v  d  n  i     l  e  M  p  r\n\r"),
			current_header->have_from);

	Write_to_screen(FRM("\n\rOffset: %-22ld%3d%3d%3d%3d%3d%3d%3d%3d%3d%3d%3d%3d%3d%3d%3d%3d\n"),
			current_header->offset,
			(current_header->status & ACTION) != 0,
			(current_header->status & CONFIDENTIAL) != 0,
			(current_header->status & DELETED) != 0,
			(current_header->status & EXPIRED) != 0,
			(current_header->status & FORM_LETTER) != 0,
			(current_header->status & NEW) != 0,
			(current_header->status & PRIVATE_MAIL) != 0,
			(current_header->status & TAGGED) != 0,
			(current_header->status & URGENT) != 0,
			(current_header->status & VISIBLE) != 0,
			(current_header->status & UNREAD) != 0,
			(current_header->status & REPLIED) != 0,
			(current_header->status & MIME_MESSAGE) != 0,
			(current_header->status & PRE_MIME_CONTENT) != 0,
			(current_header->status & MIME_UNSUPPORTED) != 0,
			(current_header->status & NOHDRENCODING) != 0);
	
        Write_to_screen(FRM("\n\rReceived on: %s\r"),
			asctime(localtime(&current_header->received_time)));
      

	header_time = current_header->time_sent + current_header->tz_offset;
	Write_to_screen(FRM("\rMessage sent on: %s\r\nFrom timezone: %s (%d)\n\r"),
			asctime(gmtime(&header_time)),
			current_header->time_zone,
			current_header->tz_offset);
	
	Write_to_screen(FRM("(Env) From: %s\n\r"), 
			current_header->env_from);

	if (current_header->from) {
	  int first = 1;
	  struct addr_item *p;
	  for (p = current_header->from; 
	       p->fullname && p->addr && p->comment; 
	       p++) {
	    if (first)
	      Write_to_screen(FRM(     "From: "));
	    else
	      Write_to_screen(FRM(",\r\n      "));
	    Write_to_screen(FRM("%-30S <%s>"),p->fullname,p->addr);
	    first = 0;
	    if (string_len(p->comment)) 
		Write_to_screen(FRM(" (%S)"),p->comment);
	  }
	  Write_to_screen(FRM("\r\n"));
	}

	

	Write_to_screen(FRM("Subject: %S\n\rInternal Index Reference Number = %d/%d (hdr index %d)\n\r"),
			current_header->subject,
			index->mailbox_number,
			index->index,
			current_header->index_number_X);

	if (current_header->to) {
	    int first = 1;
	    struct addr_item *p;
	    for (p = current_header->to; 
		 p->fullname && p->addr && p->comment; 
		 p++) {
		if (first)
		    Write_to_screen(FRM(     "To: "));
		else
		    Write_to_screen(FRM(",\r\n    "));
		Write_to_screen(FRM("%-30S <%s>"),p->fullname,p->addr);
		first = 0;
		if (string_len(p->comment)) 
		    Write_to_screen(FRM(" (%S)"),p->comment);
	    }
	    Write_to_screen(FRM("\r\n"));
	}

	if (current_header->cc) {
	    int first = 1;
	    struct addr_item *p;
	    for (p = current_header->cc; 
		 p->fullname && p->addr && p->comment; 
		 p++) {
		if (first)
		    Write_to_screen(FRM(     "CC: "));
		else
		    Write_to_screen(FRM(",\r\n    "));
		Write_to_screen(FRM("%-30S <%s>"),p->fullname,p->addr);
		first = 0;
		if (string_len(p->comment)) 
		    Write_to_screen(FRM(" (%S)"),p->comment);	    
	    }
	    Write_to_screen(FRM("\r\n"));
	}


	Write_to_screen(FRM("Message-ID: %s\n\r"), 
			strlen(current_header->messageid) > 0 ? 
			current_header->messageid : "<none>");

	Write_to_screen(FRM("Status: %s\n\r"),  
			current_header->mailx_status);

	Write_to_screen(FRM("Content-Transfer-Encoding: %s\n\r"),
			ENCODING(current_header->mime_rec.encoding));

 fail1:
	PutLineX(LINES-1,0,CATGETS(elm_msg_cat, ElmSet, 
			     ElmDbxPressAnyKey,
				     "Please Press any key to return."));
	if (REDRAW_MARK == menu_ReadCh(page, REDRAW_MARK))
	  goto redraw;

	erase_menu_context(&page);
	return;

}

static void do_check_only(to_whom,aview)
     char *to_whom[];
     struct AliasView *aview;
{
    struct expanded_address A;
    
    zero_expanded_address(&A);
    
    dump_expanded_address(3,"Check-only: checking",A);

    if (to_whom) {
	struct mailer_info *mailer_info = get_mailer_info();
	if (argv_to_expanded(&A,to_whom,mailer_info,aview)) {
	    struct string * addr_string = hdr_to_expval(A);
	    if (addr_string) {
		elm_fprintf(stdout,
			    CATGETS(elm_msg_cat, ElmSet, ElmExpandsTo,
				    "Expands to: %S\n"),
			    addr_string);
		free_string(&addr_string);
	    }
	}
	free_expanded_address(&A);
	free_mailer_info(&mailer_info);
    }
}

void check_range(mc)
     struct menu_common *mc;
{
    int i;
    int current  = mcommon_get_current(mc);
    int selected = mcommon_get_selected(mc);

    i = compute_visible(current, mc);

    if ((current < 1) || (selected && i < 1)) {
	if (mcommon_get_count(mc) > 0) {
	    /* We are out of range! Get to first message! */
	    if (selected)
		current = compute_visible(1,mc);
	    else
		current = 1;
	}
	else
	    current = 0;
    }
    else if ((current > mcommon_get_count(mc))
	      || (selected && i > selected)) {

	if (mcommon_get_count(mc) > 0) {
	    /* We are out of range! Get to last (visible) message! */
	    if (selected)
		current = visible_to_index(selected,mc)+1;
	    else
		current = mcommon_get_count(mc);

	} else
	    current = 0;
    }
    
    mcommon_set_current(mc,current);
    
}

static char *no_aliases = NULL;

/* Return key if unknown */
int motion(ch,mc, nufoot, header_page, page)
     int ch;
     struct menu_common *mc;    
     int *nufoot;
     int *header_page;
     struct menu_context  *page;
{
    /* Consolidated the standard menu navigation and delete/tag
     * commands to a function.                                   */

    int  i;
    int redraw = 0;

    int current = mcommon_get_current(mc);

	switch (ch) {


	case FIND_MARK:
	case '/'    :  /* scan mbox or aliases for string */
	    if  (mcommon_get_count(mc) < 1) {
		lib_error(CATGETS(elm_msg_cat, ElmSet,
				  ElmNoItemToScan, "No %S to scan!"), 
			  mcommon_give_item(mc,m_items));
			
	    }
	    else {
		int x = pattern_match(mc, *header_page,
				      page);
		
		redraw += menu_need_redraw(page);


		if (EOF == x) {
		    return EOF;   /* ERROR */

		}

		if (x) {
		    current = mcommon_get_current(mc);
		    nucurr = get_page(current,mc, header_page);
		} else {
		    lib_error(CATGETS(elm_msg_cat, ElmSet, 
				      ElmPatternNotFound,
				      "pattern not found!"));
		}
	    }
	    break;

	case PAGEDOWN_MARK :
	case RIGHT_MARK :
	case '+'	:  {
	    int selected = mcommon_get_selected(mc);
	    
	    /* move to next page if we're not on the last */

	    if((selected &&
		(((*header_page)+1)*headers_per_page < selected))
	       ||(!selected &&
		  (((*header_page)+1)*headers_per_page < mcommon_get_count(mc)))) {
		
		(*header_page)++;
		nucurr = NEW_PAGE;
		
		if(move_when_paged) {
		    /* move to first message of new page */
		    if(selected)
			current = visible_to_index((*header_page) * 
						   headers_per_page + 1,
						   mc) + 1;
		    else
			current = (*header_page) * headers_per_page + 1;
		    mcommon_set_current(mc,current);
		}
	    } else
		lib_error(CATGETS(elm_msg_cat, ElmSet, 
				  ElmAlreadyOnLastPage,
				  "Already on last page."));
	}
	    break;

	case PAGEUP_MARK :
	case LEFT_MARK  :
	    case '-'	:  {
		int selected = mcommon_get_selected(mc);
		
		/* move to prev page if we're not on the first */
		if((*header_page) > 0) {
		    (*header_page)--;
		    nucurr = NEW_PAGE;
		    
		    if(move_when_paged) {
			/* move to first message of new page */
			if(selected)
			    current = visible_to_index(
						       (*header_page) * headers_per_page + 1,
						       mc) + 1;
			else
			    current = (*header_page) * headers_per_page + 1;
			mcommon_set_current(mc,current);
		    }
		} else
		    lib_error(CATGETS(elm_msg_cat, ElmSet, 
				      ElmAlreadyOnFirstPage,
				      
				      "Already on first page."));
	    }
		break;

	case HOME_MARK:
	    case '='    :  {
		int selected = mcommon_get_selected(mc);

		if (selected)
		    current = visible_to_index(1,mc)+1;
		else
		    current = 1;
		mcommon_set_current(mc,current);
		nucurr = get_page(current, mc, header_page);
	    }
		break;

	    case '*'    :  {
		int selected = mcommon_get_selected(mc);

		if (selected) 
		    current = (visible_to_index(selected,mc)+1);
		else
		    current = mcommon_get_count(mc);
	
		mcommon_set_current(mc,current);
		nucurr = get_page(current, mc, header_page);
	    }
		break;

	case EOF    : return EOF;   /* ERROR */


	    case ctrl('D') :
	    case '^'    :
	case 'd'    :  
                       if (mcommon_get_count(mc) < 1) {
			   lib_error(CATGETS(elm_msg_cat, ElmSet, 
					     ElmNoItemToDelete,
					     "No %S to delete!"), 
				     mcommon_give_item(mc,m_item));
			   
		       } else {
			     if(ch == ctrl('D')) {

				 /* if current item did not become deleted,
				  * don't to move to the next undeleted item */
				 if(!meta_match(DELETED, mc, *header_page,
						page)) 
				     break;

			     } else 
				 delete_msg((ch == 'd'), TRUE, mc,
					    *header_page);

			     /* If given '^' command (toggle delete flag)
				go to next command regagdles is next command
				marked for deletion or not on resolve mode.
				In other words works same way than u)undelete
				command works.
			     */

			     if (resolve_mode) 	/* move after mail resolved */
			       if((i=next_message(current-1, 
						  (ch == 'd') ? TRUE : FALSE,
						  mc)) != -1) {
				   
				   current = i+1;
				   mcommon_set_current(mc,current);

				   nucurr = get_page(current, mc, header_page);
			       }
			   }
			   break;

	    case 'J'    :  if(current > 0) {
			     if((i=next_message(current-1, FALSE,
						mc)) != -1) {

				 DPRINT(Debug,5,(&Debug,  
						 " current %d => %d\n",
						 current,i+1));

			       current = i+1;
			       mcommon_set_current(mc,current);

			       nucurr = get_page(current,mc, header_page);
			     } else
			       lib_error(CATGETS(elm_msg_cat, ElmSet, 
						 ElmNoMoreItemBelow,
						 "No more %S below."), 
					 mcommon_give_item(mc,m_items));
	    } else 
		lib_error(FRM("%S"),
			  mcommon_give_item(mc,m_no_item));
		
		break;

next_undel_msg:
	case DOWN_MARK :
	    case 'j'    :  if(current > 0) {
			     if((i=next_message(current-1, TRUE,
						mc)) != -1) {

				 DPRINT(Debug,5,(&Debug,  
						 " current %d => %d\n",
						 current,i+1));

			       current = i+1;
			       mcommon_set_current(mc,current);

			       nucurr = get_page(current,mc, header_page);
			     } else
			       lib_error(CATGETS(elm_msg_cat, ElmSet, 
						 ElmNoItemUndeletedBelow,
						 "No more undeleted %S below."), 
					 mcommon_give_item(mc,m_items));
	    } else
		lib_error(FRM("%S"),
			  mcommon_give_item(mc,m_no_item));
		
		break;

	    case 'K'    :  if(current > 0) {
			     if((i=prev_message(current-1, FALSE,
						mc)) != -1) {

				 DPRINT(Debug,5,(&Debug,  
						 " current %d => %d\n",
						 current,i+1));

			       current = i+1;
			       mcommon_set_current(mc,current);

			       nucurr = get_page(current,mc, header_page);
			     } else
			       lib_error(CATGETS(elm_msg_cat, ElmSet, 
						 ElmNoMoreItemAbove,
						 "No more %S above."), 
					 mcommon_give_item(mc,m_items));
	    } else 
		lib_error(FRM("%S"),
			  mcommon_give_item(mc,m_no_item));
		
		break;

	case UP_MARK :
	    case 'k'    :  if(current > 0) {
			     if((i=prev_message(current-1, TRUE, 
						mc)) != -1) {

				 DPRINT(Debug,5,(&Debug,  
						 " current %d => %d\n",
						 current,i+1));

			       current = i+1;
			       mcommon_set_current(mc,current);

			       nucurr = get_page(current, mc, header_page);
			     } else
			       lib_error(CATGETS(elm_msg_cat, ElmSet, 
						 ElmNoMoreUndeletedAbove,
						 "No more undeleted %S above."), 
					 mcommon_give_item(mc,m_items));
	    } else 
		lib_error(FRM("%S"),
			  mcommon_give_item(mc,m_no_item));
		
		break;

	case 'l'    :  Write_to_screen(CATGETS(elm_msg_cat, ElmSet,
					       ElmLimitDisplayBy,
					       "Limit displayed %S by..."), 
				       mcommon_give_item(mc,m_items));
			   clear_error();
			   limit(mc, page);
			   if (menu_need_redraw(page)) {
			       current = mcommon_get_current(mc);
			       get_page(current, mc, header_page);
			       redraw++;
			   } else {
			     (*nufoot)++;
			   }
			   break;

            case ctrl('T') :
	    case 'T'	   :
	    case 't'       :  if (mcommon_get_count(mc) < 1) {
				lib_error(CATGETS(elm_msg_cat, ElmSet, 
						  ElmNoItemToTag,
						  "No %S to tag!"), 
					  mcommon_give_item(mc,m_items));
				
			      }
			      else if (ch == 't')
				  tag_message(TRUE, mc,*header_page); 
			      else if (ch == 'T') {
				  tag_message(TRUE, mc, *header_page); 
				goto next_undel_msg;
			      }
			      else
				  meta_match(TAGGED, mc, *header_page,
					     page);
			      break;


	case 'u'    :  if (mcommon_get_count(mc) < 1) {
			     lib_error(CATGETS(elm_msg_cat, ElmSet, 
					       ElmNoItemToMarkUndeleted,
					       "No %S to mark as undeleted!"), 
				       mcommon_give_item(mc,m_items));
			     
			   }
			   else {
			       undelete_msg(TRUE, mc, *header_page);
			     if (resolve_mode) 	/* move after mail resolved */
			       if((i=next_message(current-1, FALSE, mc)) != -1) {

				   
				   current = i+1;
				   mcommon_set_current(mc,current);
				   nucurr = get_page(current, mc, header_page);

			       }
/*************************************************************************
 **  What we've done here is to special case the "U)ndelete" command to
 **  ignore whether the next message is marked for deletion or not.  The
 **  reason is obvious upon usage - it's a real pain to undelete a series
 **  of messages without this quirk.  Thanks to Jim Davis @ HPLabs for
 **  suggesting this more intuitive behaviour.
 **
 **  The old way, for those people that might want to see what the previous
 **  behaviour was to call next_message with TRUE, not FALSE.
**************************************************************************/
			   }
			   break;

	    case ctrl('U') : if (mcommon_get_count(mc) < 1) {
			       lib_error(CATGETS(elm_msg_cat, ElmSet, 
						 ElmNoItemToUndelete,
						 "No %S to undelete!"), 
					 mcommon_give_item(mc,m_items));
			       
			     }
			     else 
				 meta_match(UNDELETE, mc, *header_page,
					    page);
			     break;

	    case ctrl('L') : redraw++;	break;

	    case NO_OP_COMMAND : break;	/* noop for timeout loop */

	    default	: if (ch > '0' && ch <= '9') {
		int selected = mcommon_get_selected(mc);
		
		Write_to_screen(CATGETS(elm_msg_cat, ElmSet,
					ElmNewCurrentItem,
						     "New Current %S"), 
				mcommon_give_item(mc,m_Item));

		i = read_number(ch, 
				mcommon_give_item(mc,m_item),
				current, page);

			    DPRINT(Debug,5,(&Debug,  
					    " current %d => %d (?)\n",
					    current,i));

			    if( i > mcommon_get_count(mc))
				lib_error(CATGETS(elm_msg_cat, ElmSet, 
						  ElmNotThatMany,
						  "Not that many %S."), 
					  mcommon_give_item(mc,m_items));
			    else if(selected
				    && mcommon_isoff_status(mc,i-1,VISIBLE))
				lib_error(CATGETS(elm_msg_cat, ElmSet, 
						  ElmNotInLimitedDisplay,
						  "%S not in limited display."),
					  mcommon_give_item(mc,m_Item));
			    else {
				current = i;
				mcommon_set_current(mc,current);
				
				nucurr = get_page(current, mc, header_page);
			    }
			  }
			  else {
			      if (redraw)
				  menu_trigger_redraw(page);
			      return ch;  /* UNKNOWN command */
			      
			  }
	}

	if (redraw)
	    menu_trigger_redraw(page);
	return 0;   /* COmmand OK */
}

/*
 * Local Variables:
 *  mode:c
 *  c-basic-offset:4
 *  buffer-file-coding-system: iso-8859-1
 * End:
 */
