static char rcsid[] = "@(#)$Id: pattern.c,v 1.36 2004/07/26 12:41:47 hurtta Exp $";

/******************************************************************************
 *  The Elm (ME+) Mail System  -  $Revision: 1.36 $   $State: Exp $
 *
 *  Modified by: Kari Hurtta <hurtta+elm@posti.FMI.FI> 
 *                           (was hurtta+elm@ozone.FMI.FI)
 ******************************************************************************
 *  The Elm Mail System 
 *
 *			Copyright (c) 1988-1992 USENET Community Trust
 *			Copyright (c) 1986,1987 Dave Taylor
 ****************************************************************************/

/**    General pattern matching for the ELM mailer.     

**/

#include "def_elm.h"
#include "s_elm.h"

static unsigned char *s2us P_((char *str));
static unsigned char *s2us(str) 
     char *str;
{
    return (unsigned char *)str;
}

static CONST unsigned char *cs2us P_((const char *str));
static CONST unsigned char *cs2us(str) 
     CONST char *str;
{
    return (CONST unsigned char *)str;
}

#include <errno.h>
#ifndef ANSI_C
extern int errno;
#endif       

#if ANSI_C
#define S_(x) static x;
#else
#define S_(x)
#endif

char *error_description(), *shift_lower();
static char *tag_word = NULL;
static char *tagged_word = NULL;
static char *Tagged_word = NULL;
static char *delete_word = NULL;
static char *mark_delete_word = NULL;
static char *Mark_delete_word = NULL;
static char *undelete_word = NULL;
static char *undeleted_word = NULL;
static char *Undeleted_word = NULL;

static int DoAll P_((int func,struct menu_common *menu, int header_page));

DEBUG_VAR(Debug,__FILE__,"mbox");

int meta_match(function, menu, header_page,page)
     int function;
     struct menu_common *menu;
     int header_page;
     struct menu_context *page; 
{
    char	ch, tagmsg[SLEN];
    
    /** Perform specific function based on whether an entered string 
	matches either the From or Subject lines.. 
	Return TRUE if the current message was matched, else FALSE.
    **/
    
    int i, tagged=0, count=0, curtag = 0;
    char		msg[SLEN];

    struct string * meta_pattern = NULL;
    int status;
    int LINES, COLUMNS;

    menu_get_sizes(page,&LINES, &COLUMNS);

    if (delete_word == NULL) {
	tag_word = catgets(elm_msg_cat, ElmSet, ElmTag, "Tag");
	tagged_word = catgets(elm_msg_cat, ElmSet, ElmTagged, "tagged");
	Tagged_word = catgets(elm_msg_cat, ElmSet, ElmCapTagged, "Tagged");
	delete_word = catgets(elm_msg_cat, ElmSet, ElmDelete, "Delete");
	mark_delete_word = catgets(elm_msg_cat, ElmSet, ElmMarkDelete, "marked for deletion");
	Mark_delete_word = catgets(elm_msg_cat, ElmSet, ElmCapMarkDelete, "Marked for deletion");
	undelete_word = catgets(elm_msg_cat, ElmSet, ElmUndelete, "Undelete");
	undeleted_word = catgets(elm_msg_cat, ElmSet, ElmUndeleted, "undeleted");
	Undeleted_word = catgets(elm_msg_cat, ElmSet, ElmCapUndeleted, "Undeleted");
    }
    
    Write_to_screen(CATGETS(elm_msg_cat, ElmSet, 
			    ElmMessagesMatchPattern,
			    "%s %S that match pattern..."), 
		    function==TAGGED?tag_word: function==DELETED?delete_word:undelete_word, 
	     mcommon_give_item(menu, m_items));
    
    if (function == TAGGED) {	/* are messages already tagged??? */
	int mc = mcommon_get_count(menu);
	
	for (i=0; i < mc; i++) {
	    if (mcommon_ison_status(menu,i,TAGGED))
		tagged++;
	}

	if (tagged) {
	    if (tagged > 1) {
		elm_sfprintf(tagmsg, sizeof tagmsg,
			     CATGETS(elm_msg_cat, ElmSet, 
				     ElmSomeMessagesATagged,
				     "Some %S are already tagged."), 
			     mcommon_give_item(menu,m_items));
		elm_sfprintf(msg, sizeof msg,
			     CATGETS(elm_msg_cat, ElmSet, ElmRemoveTags,
				     "%s Remove Tags? (%c/%c) "),
			     tagmsg, *def_ans_yes, *def_ans_no);
	    } else {
		elm_sfprintf(tagmsg, sizeof tagmsg,
			     CATGETS(elm_msg_cat, ElmSet, 
				     ElmAMessageATagged,
				     "One %S is already tagged."), 
			     mcommon_give_item(menu,m_item));
		elm_sfprintf(msg, sizeof msg,
			     CATGETS(elm_msg_cat, ElmSet, ElmRemoveTag,
				     "%s Remove Tag? (%c/%c) "),
			     tagmsg, *def_ans_yes, *def_ans_no);
	    }
	    
	    ch = want_to(msg, *def_ans_yes, LINES-3, 0, page);
	    if (ch != *def_ans_no) {	/* remove tags... */
		for (i=0; i < mc; i++) {
		    mcommon_clearf_status(menu,i,TAGGED);
		    show_new_status(i, menu, header_page);
		}
	    }
	}
    }
	

    status = optionally_enter2(page,
			       &meta_pattern, LINES-1-2, 0,
			       OE_REDRAW_MARK|
			       OE_SIG_CHAR /* Ctrl-C */, 
			       CATGETS(elm_msg_cat, ElmSet, ElmEnterPattern, 
				       "Enter pattern: "));

    while (status == REDRAW_MARK) {
	status = optionally_enter2(page,
				   &meta_pattern, LINES-1-2, 0,
				   OE_REDRAW_MARK|OE_APPEND_CURRENT|
				   OE_SIG_CHAR /* Ctrl-C */,
				   CATGETS(elm_msg_cat, ElmSet, 
					   ElmEnterPattern, 
					   "Enter pattern: "));
    }
    
    if (status != 0 || !meta_pattern ||
	string_len(meta_pattern) == 0) {

	if (meta_pattern)
	    free_string( & (meta_pattern));

	ClearLine(LINES-1-2);
	return(curtag);
    }
    
    if (function == DELETED && 
	string_matches_ascii(meta_pattern, s2us("tagged"))) {

	if (meta_pattern)
	    free_string( & (meta_pattern));
	
	return(DeleteTagged(menu, header_page, page));

    } else if (string_matches_ascii(meta_pattern, s2us("all")))
	count = DoAll(function, menu, header_page);

    else {
	int mc       = mcommon_get_count(menu);
	int current  = mcommon_get_current(menu);
	int selected = mcommon_get_selected(menu);

	for (i = 0; i < mc; i++) {
	    if (mcommon_matches(menu,i,meta_pattern) &&
		(!selected ||
		 (selected && mcommon_ison_status(menu,i,VISIBLE))
		 )) {
		 
		if (function == UNDELETE)
		    mcommon_clearf_status(menu,i,DELETED);

		else if ((function == DELETED) && 
			 (mcommon_get_type(menu,i) & SYSTEM)) {

		    if(i == current - 1) curtag--;
		    count--;

		} else
		    mcommon_setf_status(menu,i, function);

		show_new_status(i,menu, header_page);
		if(i == current - 1) curtag++;
		count++;
	    }
	}
    }

    ClearLine(LINES-1-2);	/* remove "pattern: " prompt */
    
    if (count > 1) {
	lib_error(CATGETS(elm_msg_cat, ElmSet, ElmTaggedMessages,
			  "%s %d %S."), 
		  function==TAGGED? Tagged_word : 
		  function==DELETED? Mark_delete_word : Undeleted_word,
		  count, 
		  mcommon_give_item(menu,m_items));
    } else if (count == 1) {
	lib_error(CATGETS(elm_msg_cat, ElmSet, ElmTaggedMessage,
			  "%s 1 %S."), 
		  function==TAGGED? Tagged_word : 
		  function==DELETED? Mark_delete_word : Undeleted_word, 
		  mcommon_give_item(menu,m_item));
	} else {
	    lib_error(CATGETS(elm_msg_cat, ElmSet, ElmNoMatchesNoTags,
			      "No matches. No %S %s."), 
		      mcommon_give_item(menu,m_items),
		      function==TAGGED? tagged_word : 
		      function==DELETED? mark_delete_word: undeleted_word);
	}
    
    if (meta_pattern)
	free_string( & (meta_pattern));

    return(curtag);
}

int pattern_match(menu, header_page, page)
     struct menu_common *menu;
     int header_page;
     struct menu_context *page;
{
    /** Get a pattern from the user and try to match it with the
	from/subject lines being displayed.  If matched (ignoring
	case), move current message pointer to that message, if
	not, error and return ZERO **/
    
    /* If redraw is needed use  
         menu_trigger_redraw(page)
    */

    int i;
    
    /* Elm 2.4 remembers old entered patterns, so we need to keep
       that static. This leaks these malloced areas on end of program */

    static struct string * pattern = NULL;
    static struct string * alt_pattern = NULL;
    int ret = 0 ;

    int redraw = 0;
    int anywhere = FALSE;
    int matched;
    struct string * buffer = NULL;
    struct string *buffer1 = NULL;

    struct string * SEARCH = NULL;  /* SHARED: point either to
				       pattern or alt_pattern */

    int val;
    int LINES, COLUMNS;

    menu_get_sizes(page,&LINES, &COLUMNS);

    buffer = format_string(CATGETS(elm_msg_cat, ElmSet,
				   ElmMatchPatternInEntire, 
				   "Match pattern (in entire %S):"),
			   mcommon_give_item(menu,m_item));

    PutLineX(LINES-4,40, CATGETS(elm_msg_cat, ElmSet, ElmMatchAnywhere,
				     "/ = Match anywhere in %S."), 
	     mcommon_give_item(menu, m_items));
	
    buffer1 = format_string(CATGETS(elm_msg_cat, ElmSet, ElmMatchPattern,
				   "Match pattern:"));



    val = pattern_enter(&pattern, &alt_pattern, LINES-2,0, 
			buffer, buffer1, &redraw, page);

    /* 0 == pattern given
       1 == alt_pattern given
       -1 == EOF ...
 
    */

    if (val < 0) {
	ret = EOF;
	goto out;
    }

    if (1 == val) {
	if (! alt_pattern ||
	    string_len(alt_pattern) == 0) {
	    
	    ret = 1;
	    goto out;
	  }

	  anywhere = TRUE;
	  SEARCH = alt_pattern;
    }

    if (0 == val) {
	if (! pattern ||
	    string_len(pattern) == 0) {
	    ret = 0;
	    goto out;
	}

	anywhere = FALSE;
	SEARCH = pattern;
    }

    if (SEARCH) {

	  if (anywhere) 
	      ret = mcommon_match_in_text(menu,SEARCH, header_page,
					  page);
	  else {

	      int current = mcommon_get_current(menu);
	      int mc = mcommon_get_count(menu);
	      int selected = mcommon_get_selected(menu);

	      for (i = current; i < mc; i++) {
		  if (mcommon_matches(menu,i,SEARCH) &&
		      (!selected ||
		       (selected && mcommon_ison_status(menu,i,VISIBLE))
		       )) {
		      
		      current = ++i;
		      mcommon_set_current(menu,current);

		      ret = 1;
		      goto out;
		  }
	      }
	  }
    }

 out:

    /* NOTE: 
       We do not free pattern and alt_pattern -- they are static
       and should have remembered, so that behauviour matches
       to Elm 2.4
    */

    if (buffer)
	free_string(&buffer);
    if (buffer1)
	free_string(&buffer1);
    
    if (redraw)
	menu_trigger_redraw(page);

    return(ret);
}

/* This will tag, delete, or undelete all or the currently visible messages */
static int DoAll(func, menu, header_page)
     int func;
     struct menu_common *menu;
     int header_page;
{
    int i, count = 0;
    int mc = mcommon_get_count(menu);
    int selected = mcommon_get_selected(menu);

    for (i=0; i < mc; i++) {

	if (!selected ||
	    (selected && mcommon_ison_status(menu,i,VISIBLE))) {
	    if (func == UNDELETE)
		mcommon_clearf_status(menu,i,DELETED);
	    else
		mcommon_setf_status(menu,i,func);

	    show_new_status(i,menu, header_page);
	    ++count;
	}
    }
    return(count);
}

/*
 * Local Variables:
 *  mode:c
 *  c-basic-offset:4
 *  buffer-file-coding-system: iso-8859-1
 * End:
 */
