static char rcsid[] = "@(#)$Id: screen.c,v 1.35 2004/03/27 18:31:46 hurtta Exp $";

/******************************************************************************
 *  The Elm (ME+) Mail System  -  $Revision: 1.35 $   $State: Exp $
 *
 *  Modified by: Kari Hurtta <hurtta+elm@posti.FMI.FI> 
 *                           (was hurtta+elm@ozone.FMI.FI)
 ******************************************************************************
 *  The Elm Mail System 
 *
 *			Copyright (c) 1988-1992 USENET Community Trust
 *			Copyright (c) 1986,1987 Dave Taylor
 *****************************************************************************/

/**  screen display routines for ELM program 

**/

#include "def_elm.h"
#include "s_elm.h"

DEBUG_VAR(Debug,__FILE__,"ui");

#define  minimum(a,b)	((a) < (b) ? (a) : (b))

extern char version_buff[];

static int fix_header_page P_((struct menu_common *menu,
			       int *header_page));

static char *us2s P_((unsigned char *str));
static char *us2s(str)
     unsigned char *str;
{
    return (char *)str;
}



void showscreen(mailbox,menu, header_page, page)
     struct MailboxView *mailbox;
     menu_func *menu;
     int *header_page;
     struct menu_context *page;   /* new page assumed */
{
    struct menu_common MENU;
    
    set_mcommon_from_mbxview(&MENU,mailbox);

    menu_ClearScreen(page);
    
    update_title(mailbox, page);

    last_header_page = -1;	 	/* force a redraw regardless */
    show_headers(&MENU, header_page, page);
    
    if (mini_menu)
	menu(page);

    show_last_error();
	
    /* define_softkeys(MAIN); */
}

void set_folder_window_title(folder, title_s, icon_s)
     struct folder_info *folder;
     char *title_s, *icon_s;
{
    /* For a moment we assume that widow titles are ascii only */

    charset_t ascii_ptr = MIME_name_to_charset("US-ASCII",0);

    struct string *title_1 = 
	format_string(CATGETS(elm_msg_cat, ElmSet, 
			      ElmWindowTitle,
			      "Elm: %s %S"),
		      folder_type(folder),
		      folder->cur_folder_disp);

    struct string *title_2;    
    char *x;
    char *icon, *title_3;

    if (!ascii_ptr)
        panic("ELM PANIC",__FILE__,__LINE__,"set_folder_window_title",
              "US-ASCII not found",0);
   
    title_2 = convert_string(ascii_ptr,title_1,0);

    /* Keep it simple ... */
    x = strrchr(folder->cur_folder_sys,'/');
    if (x)
	x++;
    else 
	x = folder->cur_folder_sys;

    icon = elm_message(CATGETS(elm_msg_cat, ElmSet, 
			       ElmWindowIcon,
			       "Elm: %s"),
		       x);

    title_3 = us2s(stream_from_string(title_2,1,NULL));

    set_terminal_titles(title_s ? title_s : title_3,
			icon_s ? icon_s : icon);

    free(title_3);
    free_string(&title_2);
    free_string(&title_1);
}


void update_title(mailbox, page)
     struct MailboxView *mailbox;
     struct menu_context  *page;
{
    /** display a new title line, probably due to new mail arriving **/
    
    struct string * buffer = NULL;
    struct string * buffer2 = NULL;
    struct string * f1 = mailbox_title(mailbox);
    int message_count  = get_message_count(mailbox);   
    int selected       = get_selected(mailbox);

    int l1,l2,l;
    int LINES, COLUMNS;

    menu_get_sizes(page, &LINES, &COLUMNS);   

    ClearLine(1);
    ClearLine(2);

    if (selected)
	buffer = format_string(CATGETS(elm_msg_cat, ElmSet, 
				       ElmShownXWithSelect,
				       "%S with %d shown out of %d"),
			       f1,selected, 
			       message_count);
    else if (message_count == 1)
	buffer = format_string(CATGETS(elm_msg_cat, ElmSet, 
				       ElmShownXNoSelect,
				       "%S with 1 message"),
			       f1);
    else
	buffer = format_string(CATGETS(elm_msg_cat, ElmSet, 
				       ElmShownXNoSelectPlural,
				       "%S with %d messages"),
			       f1,message_count);
    
    free_string(&f1);
    
    l1 = string_len(buffer);
    
    buffer2 = format_string(FRM("[ELM %s]"),version_buff);
    l2 = string_len(buffer2);
    l = l1 + l2 + 1; /* Assumed */
    
    if (l > COLUMNS) {
	if (l2 < COLUMNS)
	    PutLineX(2,(COLUMNS - l2)/2,FRM("%S"),buffer2);
	if (l1 > COLUMNS) 
	    PutLineX(1,1,FRM("%S"),buffer);
	else
	    PutLineX(1,(COLUMNS - l1)/2,FRM("%S"),buffer);
    } else {
	PutLineX(1,(COLUMNS - l)/2,FRM("%S %S"),buffer,buffer2);
    }
    
    free_string(&buffer);
    free_string(&buffer2);
    
}

void show_menu(page)
     struct menu_context *page; 
{
    /** write main system menu... **/
    int   LINES, COLUMNS;	

    menu_get_sizes(page,&LINES, &COLUMNS);

    if (user_level == 0) {	/* a rank beginner.  Give less options  */
	  print_format_center(LINES-1-7, 
			      CATGETS(elm_msg_cat, ElmSet, 
				      ElmLevel0MenuLine1,
				      "You can use any of the following commands by pressing the first character;"));
         print_format_center(LINES-1-6, 
			     CATGETS(elm_msg_cat, ElmSet, 
				     ElmLevel0MenuLine2,
				     "d)elete or u)ndelete mail,  m)ail a message,  r)eply or f)orward mail,  q)uit"));
	 print_format_center(LINES-1-5, 
			     CATGETS(elm_msg_cat, ElmSet, 
				     ElmLevel0MenuLine3,
				     "To read a message, press <return>.  j = move down, k = move up, ? = help"));
    } else {
	print_format_center(LINES-1-7, 
			    CATGETS(elm_msg_cat, ElmSet, 
				    ElmLevel1MenuLine1,
				    "|=pipe, !=shell, ?=help, <n>=set current to n, /=search pattern"));
        print_format_center(LINES-1-6, 
			    CATGETS(elm_msg_cat, ElmSet, 
				    ElmLevel1MenuLine2,
				    "a)lias, C)opy, c)hange folder, d)elete, e)dit, f)orward, g)roup reply, m)ail,"));
	print_format_center(LINES-1-5, 
			    CATGETS(elm_msg_cat, ElmSet, 
				    ElmLevel1MenuLine3,
				    "n)ext, o)ptions, p)rint, q)uit, r)eply, s)ave, t)ag, u)ndelete, or e(x)it"));
    }
}


int show_headers(menu, header_page, page)
     struct menu_common *menu;
     int *header_page;
     struct menu_context  *page;   /* not really currently used */
    
{
    /** Display page of headers (10) if present.  First check to 
	ensure that header_page is in bounds, fixing silently if not.
	If out of bounds, return zero, else return non-zero 
	Modified to only show headers that are "visible" to ze human
	person using ze program, eh?
    **/
    
    int this_msg = 0, line = 4, last = 0, last_line, 
	displayed = 0, using_to;

    int selected = mcommon_get_selected(menu);
    int LINES, COLUMNS;

    menu_get_sizes(default_context,&LINES, &COLUMNS);
       
    if (fix_header_page(menu,header_page))
	return(FALSE);
    
    if (selected) {
	if (((*header_page)*headers_per_page) > selected)
	    return(FALSE); 	/* too far! too far! */
	
	this_msg = visible_to_index((*header_page) * headers_per_page + 1,
				    menu);
	displayed = (*header_page) * headers_per_page;
	
	last = displayed+headers_per_page;
	
    } else {
	if ((*header_page) == last_header_page) 	/* nothing to do! */
	    return(FALSE);
	
	/** compute last header to display **/
	
	this_msg = (*header_page) * headers_per_page;
	last = this_msg + (headers_per_page - 1);
    }
    
    if (last >= mcommon_get_count(menu))
	last = mcommon_get_count(menu)-1;
    
    /** Okay, now let's show the header page! **/
    
    ClearLine(line);	/* Clear the top line... */
    
    MoveCursor(line, 0);	/* and move back to the top of the page... */
    
    DPRINT(Debug,15,(&Debug, 
		     "show_headers: inalias=%d, selected=%d, last=%d\n",
		     inalias,selected,last));
   
    while ((selected && displayed < last) || this_msg <= last) {
	
#if DEBUG && 0	    
	if (Debug.active > 10 && !inalias) {
	    struct addr_item * p;
	    struct header_rec * hdr = give_header(mailbox,this_msg);
	    
	    DPRINT(Debug,15,(&Debug, 
			     "-- %d: %x \n",
			     this_msg,hdr));

	    for (p =  hdr->from; p && p->addr; p++) {
		DPRINT(Debug,15,(&Debug, 
				 "-- %d:from[%d] addr='%s' comment='%S' fullname='%S'\n",
				 this_msg,
				 p - hdr->from,
				 p->addr,
				 p->comment,
				 p->fullname));		    
	    }
	    
	    for (p =  hdr->to; p && p->addr; p++) {
		DPRINT(Debug,15,(&Debug, 
				 "-- %d:  to[%d] addr='%s' comment='%S' fullname='%S'\n",
				 this_msg,
				 p - hdr->to,
				 p->addr,
				 p->comment,
				 p->fullname));		    
	    }
	    
	    for (p =  hdr->cc; p && p->addr; p++) {
		DPRINT(Debug,15,(&Debug, 
				 "-- %d:  cc[%d] addr='%s' comment='%S' fullname='%S'\n",
				 this_msg,
				 p - hdr->cc,
				 p->addr,
				 p->comment,
				 p->fullname));		    
	    }
	}


	if (Debug.active > 10 && inalias) {
	    struct alias_rec * a = give_alias(aview,this_msg);

	    DPRINT(Debug,15,(&Debug, 
			     "-- %d: %x alias='%s'\n",
			     this_msg,a, a->alias));

	}

#endif
	mcommon_screen_print_menu_line(menu,this_msg, page);
		
	if (selected) 
	    displayed++;

	CleartoEOLN();
	line++;		/* for clearing up in a sec... */
	
	if (selected) {
	    if ((this_msg = next_message(this_msg, FALSE, menu)) < 0)
		break;	/* GET OUTTA HERE! */
	    
	    /* the preceeding looks gross because we're using an INDEX
	       variable to pretend to be a "current" counter, and the 
	       current counter is always 1 greater than the actual 
	       index.  Does that make sense??
	    */
	}
	else
	    this_msg++;			/* even dumber...  */
	
    }
    
    /* clear unused lines */
    
    if (mini_menu)
	last_line = LINES-9;
    else
	last_line = LINES-5;
    
    while (line < last_line) {
	CleartoEOLN();
	NewLine();
	line++;
    }
    
    display_central_message();
    
    last_current = mcommon_get_current(menu);
    last_header_page = (*header_page);
    
    return(TRUE);
}

void show_current(menu, header_page, page)
     struct menu_common *menu;
     int *header_page;
     struct menu_context  *page;
{
    /** Show the new header, with all the usual checks **/
    
    int first = 0, last = 0, last_line, new_line, using_to;
    int selected = mcommon_get_selected(menu);

    (void) fix_header_page(menu, header_page);	/* Who cares what it does? ;-) */

    /** compute the first and last header on this page **/
    first = (*header_page) * headers_per_page + 1;
    last  = first + (headers_per_page - 1);

    /* if not a full page adjust last to be the real last */
    if (selected && last > selected)
	last = selected;
    if (!selected && last > mcommon_get_count(menu))
	last = mcommon_get_count(menu);
	
    /** okay, now let's show the pointers... **/

    /** have we changed??? **/
    if (mcommon_get_current(menu) == last_current) {
	DPRINT(Debug,5,(&Debug,"show_current: Not changed last_current=%d\n",
			last_current));
	return;
    }

    if (selected) {
	last_line = ((compute_visible(last_current, menu)-1) %
		     headers_per_page)+4;
	new_line  = ((compute_visible(mcommon_get_current(menu), 
				      menu)-1) % 
		     headers_per_page)+4;
    } else {
	last_line = ((last_current-1) % headers_per_page)+4;
	new_line  = ((mcommon_get_current(menu)
		      -1) % headers_per_page)+4;
    }
    
    if (has_highlighting && ! arrow_cursor) {
	
	/* clear last current if it's in proper range */
	if (last_current > 0) {		/* not a dummy value */
	    int vis = compute_visible(last_current, menu);

	    if (vis <= last && vis >= first) {
	    
		DPRINT(Debug,5,(&Debug,  
				"\nlast_current = %d ... clearing [1] before we add [2]\n", 
				last_current));
		DPRINT(Debug,5,(&Debug, 
				"first = %d, and last = %d\n\n",
				first, last));
		
		ClearLine(last_line);	    
		MoveCursor(last_line,0);
		
		mcommon_screen_print_menu_line(menu,last_current-1,
					       page);

	    } else {
		DPRINT(Debug,5,(&Debug,  "show_current: last_current %d not visible: %d (range: %d - %d)\n",
				last_current,vis,first,last));
	    }	

	}


	MoveCursor(new_line, 0);
	
	{
	    int current = mcommon_get_current(menu);
	    
	    mcommon_screen_print_menu_line(menu,current-1,
					   page);
	    
	    DPRINT(Debug,5,(&Debug,  "show_current: ... adding new (line %d) current=%d\n",
			    new_line,current));

	}

    } else {
	int current = mcommon_get_current(menu);

	if (on_page(last_current-1, menu,*header_page)) 
	    PutLine0(last_line,0,"  ");	/* remove old pointer... */
	if (on_page(current-1, menu,*header_page))
	    PutLine0(new_line, 0,"->");
    }

    last_current = mcommon_get_current(menu);

}

static int fix_header_page(menu, header_page)
     struct menu_common *menu;
     int *header_page;
{
    /** this routine will check and ensure that the current header
	page being displayed contains messages!  It will silently
	fix 'header-page' if wrong.  Returns TRUE if changed.  **/
    
    int last_page, old_header;

    old_header = (*header_page);

    last_page = (int) ((mcommon_get_count(menu)-1) / 
			   headers_per_page);
 
    if ((*header_page) > last_page) 
	(*header_page) = last_page;
    if ((*header_page) < 0) 
	(*header_page) = 0;
    
    return(old_header != (*header_page));
}

int on_page(message,menu,header_page)
     int message;
     struct menu_common *menu;
     int header_page;
{
    int selected = mcommon_get_selected(menu);

    /** Returns true iff the specified message is on the displayed page. **/

    if (selected) 
	message = compute_visible(message, menu);

    if (message >= (header_page) * headers_per_page)
	if (message < (((header_page)+1) * headers_per_page))
	    return(TRUE);
    
    return(FALSE);
}

/*
 * Local Variables:
 *  mode:c
 *  c-basic-offset:4
 *  buffer-file-coding-system: iso-8859-1
 * End:
 */
