static char rcsid[] = "@(#)$Id: context.c,v 1.5 2004/07/26 12:41:47 hurtta Exp $";

/******************************************************************************
 *  The Elm (ME+) Mail System  -  $Revision: 1.5 $   $State: Exp $
 *
 *  Author: Kari Hurtta <hurtta+elm@posti.FMI.FI>
 *****************************************************************************/

#include "def_screen.h"

DEBUG_VAR(Debug,__FILE__,"screen");

struct menu_context ROOT_MENU = {
    MENU_CONTEXT_magic,

    0,0,        /* lines, columns  */
    0,0,        /* changed, redraw */


    NULL        /* prev */,
    NULL        /* next */
};


struct menu_context COOKED_MENU = {
    MENU_CONTEXT_magic,

    0,0,        /* lines, columns  */
    0,0,        /* changed, redraw */


    NULL        /* prev */,
    NULL        /* next */
};


struct menu_context  * default_context = &COOKED_MENU;

struct menu_context  * new_menu_context()
{
    struct menu_context  *ret = safe_malloc(sizeof (*ret));

    /* bzero is defined hdrs/defs.h */
    bzero((void *)ret,sizeof (*ret));

    ret->magic      = MENU_CONTEXT_magic;


    ret->lines      = ROOT_MENU.lines;
    ret->columns    = ROOT_MENU.columns;
    ret->changed    = ROOT_MENU.changed;
    ret->redraw     = ROOT_MENU.redraw;

    /* chain to list */

    ret->next_menu       = ROOT_MENU.next_menu;
    ROOT_MENU.next_menu  = ret;
    ret->prev_menu       = &ROOT_MENU;

    if (ret->next_menu) 
	ret->next_menu->prev_menu = ret;

    if (default_context == &COOKED_MENU) {
	DPRINT(Debug,1, (&Debug, 
			 "new_menu_context: On COOKED MENU, not changing default context\n"));
    } else
	default_context = ret;

    return ret;
}

void  erase_menu_context(ptr)
     struct menu_context  **ptr;
{
    if ((*ptr)->magic != MENU_CONTEXT_magic)
	panic("SCREEN PANIC",__FILE__,__LINE__,"erase_menu_context",
	      "Bad type magic number",0);
    
    if (&ROOT_MENU == *ptr)
	panic("SCREEN PANIC",__FILE__,__LINE__,"erase_menu_context",
	      "attempt to erase root menu",0);

    if (&COOKED_MENU == *ptr)
	panic("SCREEN PANIC",__FILE__,__LINE__,"erase_menu_context",
	      "attempt to erase cooked menu",0);
	
    if (default_context == (*ptr))
	default_context = (*ptr)->prev_menu;

    if (!default_context)
	panic("SCREEN PANIC",__FILE__,__LINE__,"erase_menu_context",
	      "no default content",0);
	
    if ((*ptr)->next_menu)
	(*ptr)->next_menu->prev_menu   = (*ptr)->prev_menu;
    if ((*ptr)->prev_menu)
	(*ptr)->prev_menu->next_menu   = (*ptr)->next_menu;

    (*ptr)->magic = 0;         /* Invalidate */
    free(*ptr);
    *ptr = NULL;
}

int   menu_resized(ptr)
     struct menu_context *ptr;
{
    if (ptr->magic != MENU_CONTEXT_magic)
	panic("SCREEN PANIC",__FILE__,__LINE__,"menu_resized",
	      "Bad type magic number",0);

    check_changes();

    if (ptr->changed) {
	
	ptr->lines      = ROOT_MENU.lines;
	ptr->columns    = ROOT_MENU.columns;
	ptr->changed    = ROOT_MENU.changed;   /* ? ? ? ? */

	return 1;
    }

    return 0;
}

extern void  menu_get_sizes(ptr,lines,columns)
     struct menu_context *ptr;
     int *lines; 
     int *columns;
{
    if (ptr->magic != MENU_CONTEXT_magic)
	panic("SCREEN PANIC",__FILE__,__LINE__,"menu_get_sizes",
	      "Bad type magic number",0);

    *lines    = ptr->lines;
    *columns  = ptr->columns;
}

int   menu_need_redraw(ptr) 
     struct menu_context *ptr;
{
    if (ptr->magic != MENU_CONTEXT_magic)
	panic("SCREEN PANIC",__FILE__,__LINE__,"menu_need_redraw",
	      "Bad type magic number",0);

    check_changes();

    if (ptr->redraw) {

	/* Reset ptr->redraw ....  */

	ptr->redraw    = ROOT_MENU.redraw;   /* ? ? ? ? */

	DPRINT(Debug,7, (&Debug, 
			 "menu_need_redraw(%p): returning redraw\n",
			 ptr));

	return 1;
    }

    return 0;
}

void menu_trigger_redraw(ptr) 
     struct menu_context *ptr;
{
    if (ptr->magic != MENU_CONTEXT_magic)
	panic("SCREEN PANIC",__FILE__,__LINE__,"menu_trigger_redraw",
	      "Bad type magic number",0);
    DPRINT(Debug,7, (&Debug, 
		     "menu_trigger_redraw(%p): triggering redraw\n",
		     ptr));
    ptr->redraw = 1;
}


static VOLATILE int resize_screen = 0;	/** SIGWINCH occured?   **/
static VOLATILE int redraw_screen = 0;	/** Elm suspended?   **/


/* called from signal handler */
void menu_context_resize() 
{

    resize_screen = 1;
    redraw_screen = 1;

    ROOT_MENU.changed = 1;   /* Indicate that ROOT_MENU is invalid
				not really needed */
}

/* called from signal handler */
void menu_context_redraw() 
{

    redraw_screen = 1;
    ROOT_MENU.redraw = 1;     /* not really needed */
}

void set_root_menu(lines,columns)
     int lines; 
     int columns;
{
    int size = 0;

    if (lines != ROOT_MENU.lines &&
	-1 != lines) {
	ROOT_MENU.lines   = lines;
	COOKED_MENU.lines = lines;

	size ++;
    }

    
    if (columns != ROOT_MENU.columns &&
	-1 != columns) {
	ROOT_MENU.columns    = columns;
	COOKED_MENU.columns  = columns;

	size++;
    }

    if (size) {
	struct menu_context *ptr;

	for (ptr = ROOT_MENU.next_menu; ptr; ptr = ptr->next_menu) {

	    if (ptr->magic != MENU_CONTEXT_magic)
		panic("SCREEN PANIC",__FILE__,__LINE__,"set_root_menu",
		      "Bad type magic number",0);

	    ptr->changed = 1;
	}    
    }

}

void check_changes()
{

    if (resize_screen || redraw_screen) {
	int size = resize_screen;
	int draw = redraw_screen;

	struct menu_context *ptr;

	int lines, columns;

	resize_screen = 0;
	redraw_screen = 0;
	
	/* Real lines .. not lines-1 !!! */
	cur_ScreenSize(&lines,&columns);

	if (lines > 0 && columns > 0) {

	    /* Check that is size changed when
	       Elm ME+ was suspended
	    */
	    if (lines   != ROOT_MENU.lines ||
		columns != ROOT_MENU.columns)
		size++;
	    
	    ROOT_MENU.lines   = lines;
	    ROOT_MENU.columns = columns;
	    ROOT_MENU.changed = 0;       /* Values valid */
	    
	    COOKED_MENU.lines   = lines;
	    COOKED_MENU.columns = columns;

	} else {
	    	DPRINT(Debug,1, (&Debug, 
				 "check_changes: cur_ScreenSize returned bad sizes: lines=%d columns=%d\n",
				 lines,columns));
	}

	if (size && default_context == &COOKED_MENU)
	    COOKED_MENU.changed = 1;

	ROOT_MENU.redraw  = draw;    /* ? ? ? */

	if (draw && default_context == &COOKED_MENU)
	    COOKED_MENU.redraw = 1;

	for (ptr = ROOT_MENU.next_menu; ptr; ptr = ptr->next_menu) {

	    if (ptr->magic != MENU_CONTEXT_magic)
		panic("SCREEN PANIC",__FILE__,__LINE__,"check_changes",
		      "Bad type magic number",0);

	    if (size)
		ptr->changed = 1;
	    if (draw)
		ptr->redraw  = 1;
	}
    }
}

/*
 * Local Variables:
 *  mode:c
 *  c-basic-offset:4
 *  buffer-file-coding-system: iso-8859-1
 * End:
 */
