static char rcsid[] = "@(#)$Id: showmsg.c,v 1.38 2004/07/11 11:36:09 hurtta Exp $";

/******************************************************************************
 *  The Elm (ME+) Mail System  -  $Revision: 1.38 $   $State: Exp $
 *
 *  Modified by: Kari Hurtta <hurtta+elm@posti.FMI.FI> 
 *                           (was hurtta+elm@ozone.FMI.FI)
 ******************************************************************************
 *  The Elm Mail System 
 *
 * 			Copyright (c) 1988-1992 USENET Community Trust
 * 			Copyright (c) 1986,1987 Dave Taylor
 *****************************************************************************/

/** This file contains all the routines needed to display the specified
    message.
**/

#include "def_elm.h"
#include "s_elm.h"

DEBUG_VAR(Debug,__FILE__,"ui");

#include <errno.h>
#ifndef ANSI_C
extern int errno;
#endif       

VOLATILE int pipe_abort = FALSE; /* not used anymore, but won't compile without it */



void OverrideCharset(mailbox, header_page, page)
     struct MailboxView *mailbox;
     int header_page;
     struct menu_context  *page;
{
    int oldstat = 0, tagged = 0;
    int current              = get_current(mailbox);
    struct header_rec * chdr = give_header(mailbox,current-1);

    char override_charset[40];
    int code;
    charset_t res = NULL;
    int mc,i;
    int LINES, COLUMNS;

    menu_get_sizes(page, &LINES, &COLUMNS);   

    if (chdr) 
	oldstat = chdr->status;	/* remember */

    mc = get_message_count(mailbox);
    for (i=0; i < mc; i++) {
	struct header_rec * hdr = give_header(mailbox,i);
	if (hdr && ison(hdr->status, TAGGED)) {
	    tagged++;
	}
    }

    if (!tagged && chdr && chdr->override_charset &&
	chdr->override_charset->MIME_name)
	strfcpy(override_charset,chdr->override_charset->MIME_name,
		sizeof override_charset);
    else if (display_charset->MIME_name)
	strfcpy(override_charset,display_charset->MIME_name,
		sizeof override_charset);
    else
	strfcpy(override_charset,"none", sizeof override_charset);

 redraw:
    if (tagged > 1) 
	print_format_center(LINES-3,
                            CATGETS(elm_msg_cat, ElmSet,
                                    ElmOverrideTagged,
				    "Override charset of %d tagged messages. Use \"none\" to cancel override."),
			    tagged);
    else if (1 == tagged) 
	     print_format_center(LINES-3,
                            CATGETS(elm_msg_cat, ElmSet,
                                    ElmOverrideTaggedOne,
				    "Override charset of tagged message. Use \"none\" to cancel override."));
    else if (chdr && chdr->override_charset)
	print_format_center(LINES-3,
                            CATGETS(elm_msg_cat, ElmSet,
                                    ElmOverrideCurrentOver,
				    "Override charset of current message. Use \"none\" to cancel override."));
    else
	print_format_center(LINES-3,
                            CATGETS(elm_msg_cat, ElmSet,
                                    ElmOverrideCurrent,
				    "Override charset of current message."));

    PutLineX(LINES-2,0,CATGETS(elm_msg_cat, ElmSet, ElmOverrideCSprompt,
				   "Override charset: "));

    code = optionally_enter(override_charset,-1,-1,OE_REDRAW_MARK|
			    OE_SIG_CHAR /* Ctrl-C */,
			    sizeof override_charset, page);

    switch(code) {
    case REDRAW_MARK:
	goto redraw;
    case -1:    /* Interrupt */
	return;
    case 1:     /* EOF */
	return;
    case 0:
	/* OK */
	break;
    default:
	DPRINT(Debug,9,(&Debug, "optionally enter returned %d (unexpected)n",
			code));
	return;
    }

    if (! override_charset[0])
	return;

    if (0 == strcmp(override_charset,"none"))
	res = NULL;
    else {
	int j;

	res = MIME_name_to_charset(override_charset,0);

	if (!res) {
	    lib_error(CATGETS(elm_msg_cat, ElmSet, ElmUnknowCharset,
			      "Charset %s is unknown."),
		      override_charset);
	    
	    goto redraw;
	}

	j = charset_properties(res);
	
	if (! (j & CS_printable) ||
	    ! (j & CS_mapping)) {
	    lib_error(CATGETS(elm_msg_cat, ElmSet, ElmCharsetBadOverride,
			      "Can't use charset %s for override."),
		      override_charset);
	    goto redraw;
	}


    }


    if (tagged == 0) {

	tagged = 1;
	if (chdr)
	    setit(chdr->status, TAGGED);
    }

    mc = get_message_count(mailbox);
    for (i=0; i < mc; i++) {	/* get each tagged msg */
	struct header_rec * hdr = give_header(mailbox,i);
	if (hdr && ison(hdr->status, TAGGED)) {
	    struct menu_common MENU;
    
	    set_mcommon_from_mbxview(&MENU,mailbox);

	    hdr->override_charset = res;

	    /* If charset is overrided, it effects classification */
	    if (hdr->mime_parsed)
		mime_classify_media(&(hdr->mime_rec),hdr);
	    
	    clearit(hdr->status, TAGGED);
	    show_new_status(i,&MENU,header_page);	/* update screen, if needed */


	}
    }

    if (chdr)
	chdr->status = oldstat;	/* BACK! */

}


/* Returns 0 on quit */

struct prompt_list {
    struct mimeinfo *structure;
    char   * printable_command;
    int    selected;
    int    col;
};

static void update_prompt P_((struct prompt_list  ** list, int  *list_len,
			      int *current_offset, int idx,
			      struct menu_context  *page));

static void update_prompt (list,list_len,current_offset,idx,page)
     struct prompt_list  ** list; 
     int  *list_len;
     int *current_offset;
     int idx;
     struct menu_context  *page; /* new context assumed */
{
    int LINES, COLUMNS;

    menu_get_sizes(page,&LINES, &COLUMNS);



    if (idx < 0) {
	int i;

	*current_offset = 0;
	menu_ClearScreen(page);

	print_format_center(1,
			    CATGETS(elm_msg_cat, ElmSet, ElmMailcapMenu,
				    "Mailcap program selection"));

	for (i = *current_offset; 
	     i >= 0 && 
		 i < *current_offset - 2 + LINES / 3  &&
		 i < *list_len;
	     i++)
	    update_prompt(list,list_len,current_offset,i, page);
		 
    } else if (idx >= 0 &&
	       idx >= *current_offset &&
	       idx < *current_offset - 2 + LINES / 3 &&
	       idx < *list_len
	       ) {
	
	int X = (idx - *current_offset) * 3 + 3;
	media_type_t            T =  (*list)[idx].structure->TYPE;

	struct string * S = 
	    format_string(CATGETS(elm_msg_cat, ElmSet, ElmUseMailcap,
			 "%2d Use '%s' to show %s/%s"),
			  idx,
			  (*list)[idx].printable_command ? 
			  (*list)[idx].printable_command : "???",
			  get_major_type_name(T), get_subtype_name(T));
	int pos = 0;
	int visible_len = 0;
	struct string * S1 = curses_printable_clip(S,&pos,string_len(S),
						   &visible_len,
						   COLUMNS-10);

	PutLineX(X,0,FRM("%S ? "),S1);
	CleartoEOLN();
		 
	(*list)[idx].col = visible_len+3;
	
	if ((*list)[idx].structure->handler_data->use_entry ||
	    (*list)[idx].selected) {
	    
	    if ((*list)[idx].structure->handler_data->use_entry)
		PutLineX(X,(*list)[idx].col,
			 CATGETS(elm_msg_cat, ElmSet, ElmYesWord, 
				 "Yes."));
	    else
		PutLineX(X,(*list)[idx].col,
			 CATGETS(elm_msg_cat, ElmSet, ElmNoWord, 
				 "No."));
	}
	    
	free_string(&S1);
	free_string(&S);
    }

}

static void prompt_programs_1 P_((struct mimeinfo *structure,
				 struct prompt_list  ** list, 
				 int  *list_len,
				 int *current_offset,
				  struct menu_context  *page));

static void prompt_programs_1(structure, list, list_len, current_offset,
			      page)
     struct mimeinfo *structure;
     struct prompt_list  ** list; 
     int  *list_len;
     int *current_offset;
     struct menu_context  *page;
{
    int LINES, COLUMNS;

    menu_get_sizes(page, &LINES, &COLUMNS);   


    if (structure->mime_flags) {
       
	if (structure->handler_data &&
	    structure->handler_data->entry) {
	    
	    char * cmd = mailcap_view_command(structure->handler_data->entry,
					      structure);

	    if (cmd) {
		
		char **  trusted_programs = 
		    give_dt_path_as_elems(&internal_mailcap_t_programs,
					  "internal-mailcap-trusted-programs");
		
		int IS_OK = 0;

		int i;
		char *x;

		if (trusted_programs) {
		    for (i = 0; trusted_programs[i]; i++) {
			if (0 == strcmp(cmd,
					trusted_programs[i]))
			    IS_OK++;
		     }
		}
		*list = safe_realloc(*list,
				     sizeof ((*list)[0]) * (*list_len +1));
		
		bzero((void *) &( (*list)[*list_len] ),
		      sizeof    ( (*list)[*list_len] ));
		
		if (strlen(cmd) > 15 && (x = strrchr(cmd,'/')) && x != cmd) {

		    char * s = elm_message(FRM("...%s"),x);

		    free(cmd);
		    cmd = s;
		}

		(*list)[*list_len].structure          = structure;
		(*list)[*list_len].printable_command  = cmd;
		(*list)[*list_len].selected           = 0;
		(*list)[*list_len].col                = COLUMNS-7;

		if (IS_OK) {
		    (*list)[*list_len].selected           = 1;
		    (*list)[*list_len].structure->handler_data->use_entry = 1;
		}
		
		i = (*list_len)++;

		update_prompt(list,list_len,current_offset,i,page);
	    }
	}
	
	/* Is alternative -- just process it */
	if (structure->handler_data &&
	    structure->handler_data->selected_alternative) {

	    if (structure->handler_data->selected_alternative->magic != MIME_magic)
		mime_panic(__FILE__,__LINE__,"prompt_programs_1",
			   "Bad magic number (alternatice)");

	    prompt_programs_1(structure->handler_data->selected_alternative,
			      list,list_len,current_offset, page);

	} 

	/* Process subparts recursively */
	else if (structure->parser_data) {
	    int count = mime_parser_subparts(structure->parser_data);
	    int i;

	    for (i = 0; i < count; i++) {
		mime_t *att = 	mime_parser_index(structure->parser_data,i);

		if (att->magic != MIME_magic)
		    mime_panic(__FILE__,__LINE__,"prompt_programs_1",
			       "Bad magic number (subpart)");

		prompt_programs_1(att,
				  list,list_len,current_offset, page);
	    
	    }

	}
	
    }
}

/* return TRUE if metamail is used instead */
static int prompt_programs P_((struct mimeinfo *structure));
static int prompt_programs(structure)
     struct mimeinfo *structure;
{
    int r = TRUE;

    struct prompt_list  * list     = NULL;
    int                   list_len = 0;
    int                   offset   = 0;

    struct menu_context  *page = new_menu_context();
    int LINES, COLUMNS;

    menu_get_sizes(page, &LINES, &COLUMNS);   


    update_prompt(&list,&list_len,&offset,-1, page);
    
    prompt_programs_1(structure,&list,&list_len,&offset, page);

    if (list_len > 0) {
	int q = 0;
	int ch;

	/* Check if all programs was trusted ... */

	while (q < list_len &&
	       list[q].selected &&
	       list[q].structure &&
	       list[q].structure->handler_data &&
	       list[q].structure->handler_data->use_entry)
	    q++;

	r = FALSE;
	if (q >= list_len && !internal_mailcap_t_prompt) 
	    goto out;

	PutLineX(LINES-2,1,
		 CATGETS(elm_msg_cat, ElmSet, ElmMailcapQuit,
			 "Answer questions, use m)etamail or q)uit menu: "));

	do {
	    int defans;

	resize_mark:
	    defans = *def_ans_no;

	    if (menu_resized(page)) {
		menu_get_sizes(page, &LINES, &COLUMNS);   
	    }
	    
	    if (menu_need_redraw(page))
		goto redraw;

	    if (q >= list_len) {

		PutLineX(LINES-2,1,
			 CATGETS(elm_msg_cat, ElmSet, ElmMailcapQuit,
				 "Answer questions, use m)etamail or q)uit menu: "));
		defans = 'q';

	    } else {	       		
		int X = (q - offset) * 3 + 3;

		if (X <= 1) {
		    offset = q-1;
		    if (offset < 0)
			offset = 0;
		    ch = REDRAW_MARK;
		    goto redraw;
		}

		if (X >= LINES-2) {
		    offset = q;
		    ch = REDRAW_MARK;
		    goto redraw;
		}

		MoveCursor(X,list[q].col);
	    }

	    
	    CleartoEOLN();   /* Clear current answer */
	    
	    Write_to_screen(FRM("%c%c"),defans,BACKSPACE);
	    FlushBuffer();


	    ch = menu_ReadCh(page,REDRAW_MARK|READCH_CURSOR|READCH_resize);
	    
	    if (ch == '\n') 
		ch = defans;
	    
	    switch (ch) {
	    case RESIZE_MARK:

		DPRINT(Debug,4, (&Debug,"    .... resizing\n"));
		goto resize_mark;

	    case REDRAW_MARK:
	    redraw:
		update_prompt(&list,&list_len,&offset,-1, page);
		
		PutLineX(LINES-2,1,
			 CATGETS(elm_msg_cat, ElmSet, ElmMailcapQuit,
				 "Answer questions, use m)etamail or q)uit menu: "));
		break;
		
	    case '-':
	    case LEFT_MARK:
	    case PAGEUP_MARK:
		offset -= LINES / 3 - 2;
		if (offset < 0)
		    offset = 0;
		q = offset;
		goto redraw;

	    case '+':
	    case RIGHT_MARK:
	    case PAGEDOWN_MARK:

		offset += LINES / 3 - 2;
		if (offset >= list_len);
		    offset = list_len-1;
		q = offset;
		goto redraw;

	    case DOWN_MARK:
		if (q < list_len) {
		    update_prompt(&list,&list_len,&offset,q, page);
		    q++;
		}
		break;

	    case UP_MARK:
		if (q < list_len) 
		    update_prompt(&list,&list_len,&offset,q, page);
		if (q > 0)
		    q--;
		break;

	    case 'q':
		ch = 0;

		if (q < list_len) 
		    update_prompt(&list,&list_len,&offset,q, page);

		PutLineX(LINES-2,1,
			 CATGETS(elm_msg_cat, ElmSet, ElmMailcapQuit,
				 "Answer questions, use m)etamail or q)uit menu: "));

		Write_to_screen(CATGETS(elm_msg_cat, ElmSet,
					ElmQuit,
					"Quit"));

		break;

	    case 'm':

		if (q < list_len) 
		    update_prompt(&list,&list_len,&offset,q, page);

		PutLineX(LINES-2,1,
			 CATGETS(elm_msg_cat, ElmSet, ElmMailcapQuit,
				 "Answer questions, use m)etamail or q)uit menu: "));
		
		Write_to_screen(CATGETS(elm_msg_cat, ElmSet,
					ElmMetamail,
					"Metamail"));

		if (have_metamail()) {
		    ch = 0;
		    r = TRUE;
		}

		break;
	    }

	    if (q < list_len) {
		
		if (ch == *def_ans_yes) {
		    list[q].selected = 1;
		    list[q].structure->handler_data->use_entry = 1;
		    update_prompt(&list,&list_len,&offset,q, page);
		    q++;
		} else if (ch == *def_ans_no) {
		    
		    list[q].selected = 1;
		    list[q].structure->handler_data->use_entry = 0;
		    update_prompt(&list,&list_len,&offset,q, page);
		    q++;

		}
	    } 

	    FlushBuffer();
		
	} while(ch);
    }


 out:
 failure:

    if (list) {
	int i;

	for (i = 0; i < list_len; i++) {
	    if (list[i].printable_command) {
		free(list[i].printable_command);
		list[i].printable_command = NULL;
	    }
	}

	free(list);
	list = NULL;
    }
    list_len = 0;

    erase_menu_context(&page);

    return r;
}


int need_meta (hdr)
     struct header_rec *hdr;
{
    /* Determine whether or not we need to call metamail to display the
     * message contents.
     */
    int result = 0;

    if ((hdr->status & (MIME_MESSAGE)) && (hdr->status & MIME_UNSUPPORTED))
	result = have_metamail();

    /* Do not call metamail if Disposition is not inline */
    else if ((hdr->status & MIME_MESSAGE) && 
	     hdr->mime_rec.disposition != DISP_INLINE)
	result = FALSE;  
    else if (hdr->status & (MIME_MESSAGE | PRE_MIME_CONTENT)) {

	int X = hdr->mime_rec.mime_flags;  /* NOTPLAIN_need_metamail           0x01
					      NOTPLAIN_need_mailcap            0x02
					      NOTPLAIN_canuse_mailcap          0x04
					   */
		
	if (0 != (X & NOTPLAIN_need_metamail) && have_metamail())
	    result = TRUE;
	else if (0 != (X & NOTPLAIN_canuse_mailcap) ||
		 0 != (X & NOTPLAIN_need_mailcap)) {

	    result = prompt_programs (& (hdr->mime_rec));

	}

    }

    DPRINT(Debug,9,(&Debug, "need_meta=%d\n",result));
    return result;
}

int show_msg(current_header, infile, current,message_count, page)
     struct header_rec *current_header;
     FILE *infile;
     int current, message_count;
     struct menu_context  *page;
{
	/*** Display number'th message.  Get starting and ending lines
	     of message from headers data structure, then fly through
	     the file, displaying only those lines that are between the
	     two!

	     Return 0 to return to the index screen or a character entered
	     by the user to initiate a command without returning to
	     the index screen (to be processed via process_showmsg_cmd()).
	***/



	int is_need_meta = 0;
	int LINES, COLUMNS;
	
	menu_get_sizes(page,&LINES, &COLUMNS);

	DPRINT(Debug,4,(&Debug, 
			"displaying %d lines from message #%d\n", 
			current_header->lines, 
			current));


	if(ison(current_header->status, NEW)) {
	    clearit(current_header->status, NEW);   /* it's been read now! */
	    current_header->status_chgd = TRUE;
	}
	if(ison(current_header->status, UNREAD)) {
	  clearit(current_header->status, UNREAD);   /* it's been read now! */
	  current_header->status_chgd = TRUE;
	}



	is_need_meta = need_meta(current_header);

        if (is_need_meta) {
	        char fname[STRING], Cmd[SLEN];
		int err;
		FILE *fpout;

		elm_sfprintf(fname, sizeof fname,
			     FRM("%semm.%d.%d"), temp_dir, getpid(), getuid());

		if ((fpout = safeopen_rdwr(fname)) == NULL) {
		    err = errno;
		    DPRINT(Debug,1,(&Debug, 
				    "Error: open of temporary file %s, errno %s (show_message)\n",
				    fname, error_description(err)));
		  lib_error(CATGETS(elm_msg_cat, ElmSet, 
				    ElmCantOpenAsOutputFile,
				    "Can't open \"%s\" as output file! (%s)."),
			    fname, error_description(err));
		  return(0);
		}
		/* Let metamail decode it! 
		 * (Now CM_DECODE also decodes MIME and PGP) -KEH
		 *
		 * CM_LF: Metamail can not cope with CRLF
		 */
		if (!copy_message_f(infile,current_header,
				    "", fpout, CM_LF, NULL)) {
		    /* FAIL */
		    fclose (fpout);
		    unlink (fname);
		    return 0;
		}
		(void) fclose (fpout);
		elm_sfprintf(Cmd, sizeof Cmd,
			     FRM("%s -p -z -m Elm %s"), 
			     metamail_path, fname);
		Raw(OFF);
		ClearScreen();
		printf ("Executing metamail...\n");
		system_call(Cmd, SY_ENAB_SIGINT|SY_ENV_METAMAIL, NULL);
		(void) unlink (fname);

		if (prompt_after_metamail) {
		    struct menu_context *cpage;

		    cpage = Raw(ON | NO_TITE);	/* Raw on but don't switch screen */
		redraw:
		    menu_get_sizes(cpage,&LINES, &COLUMNS);

		    PutLineX(LINES-1,0, 
			     CATGETS(elm_msg_cat, ElmSet, 
				     ElmPressAnyKeyIndex,
				     "Press any key to return to index."));
		    if (menu_ReadCh(cpage, REDRAW_MARK) == REDRAW_MARK)
			goto redraw;

		    printf("\r\n");
		    Raw(OFF | NO_TITE); /* Raw off so raw on takes effect */
		}

		Raw(ON); /* Finally raw on and switch screen */
		return(0);
	} else {
	    
	    return (metapager (infile, current_header, TRUE, 
			       current, message_count, page));
	}
}

/*
 * Local Variables:
 *  mode:c
 *  c-basic-offset:4
 *  buffer-file-coding-system: iso-8859-1
 * End:
 */
