static char rcsid[] = "@(#)$Id: showmsg_c.c,v 1.37 2004/07/26 12:41:47 hurtta Exp $";

/******************************************************************************
 *  The Elm (ME+) Mail System  -  $Revision: 1.37 $   $State: Exp $
 *
 *  Modified by: Kari Hurtta <hurtta+elm@posti.FMI.FI> 
 *                           (was hurtta+elm@ozone.FMI.FI)
 ******************************************************************************
 *  The Elm Mail System 
 *
 *			Copyright (c) 1988-1992 USENET Community Trust
 *			Copyright (c) 1986,1987 Dave Taylor
 *****************************************************************************/

/** This is an interface for the showmsg command line.  The possible
    functions that could be invoked from the showmsg command line are
    almost as numerous as those from the main command line and include
    the following;

	   |    = pipe this message to command...
	   !    = call Unix command
	   <    = scan message for calendar info
	   b    = bounce (remail) message
	   d    = mark message for deletion
	   f    = forward message
	   g    = group reply
	   h    = redisplay this message from line #1, showing headers
	   <CR> = redisplay this message from line #1, weeding out headers
	   i,q  = move back to the index page (simply returns from function)
	   J    = move to body of next message
	   j,n  = move to body of next undeleted message
	   K    = move to body of previous message
	   k    = move to body of previous undeleted message
	   m    = mail a message out to someone
	   N    = move to body of next unread message
	   p    = print this (all tagged) message
	   r    = reply to this message
	   s    = save this message to a maibox/folder 
	   t    = tag this message
	   u    = undelete message
           v    = view attachments
	   x    = Exit Elm NOW 

    all commands not explicitly listed here are beeped at.  Use i)ndex
    to get back to the main index page, please.

    This function returns when it is ready to go back to the index
    page.
**/

#include "def_elm.h"
#include "s_elm.h"

DEBUG_VAR(Debug,__FILE__,"ELM");

int	screen_mangled = 0;
char    msg_line[SLEN];

#define store_msg(a)	strfcpy(msg_line,a, sizeof msg_line)



int showmsg_main_cmd(mailbox,cmd,screen_mangled, page)
     struct MailboxView *mailbox;
     int cmd;
     int *screen_mangled;
     struct menu_context  *page;
{
    struct menu_common MENU;
    
    set_mcommon_from_mbxview(&MENU,mailbox);

    switch (cmd) {
    case '?' : 
	
	help(TRUE, page);
	    
	if (menu_need_redraw(page)) {
	    menu_ClearScreen(page);
	    build_bottom(&MENU,page);
	} else 
	    *screen_mangled = TRUE;
	return 256;

    case '|' : {
	int current = get_current(mailbox);
	
	put_cmd_name(catgets(elm_msg_cat, ElmSet, ElmPipe, "Pipe"), 
		     TRUE, &MENU, page);
	
	if (give_message_data(mailbox,current-1,
			      NULL,NULL,NULL,
			      NO_mime_parse))
	    do_pipe(mailbox, page);  
	else {
	    DPRINT(Debug,3,(&Debug, 
			    "give_message_data [%d] fails",current-1));
	}
	menu_ClearScreen(page);
	build_bottom(&MENU, page);
    }
	return 256;

    }

    return cmd;

}


int process_showmsg_cmd(command, mailbox, XX, aview, header_page, page)
     int command;
     struct MailboxView *mailbox;
     showmsg_cmd   XX;
     struct AliasView *aview;
     int *header_page;
     struct menu_context  *page;
{
    int     i, intbuf;		/* for dummy parameters...etc */
    int	istagged;		/* for tagging and subsequent msg */
    

    /* FIXME:   Should this allocate new page context ? ? ?  */

    /* ? ? ? Why this Raw is on here ? ? ? */
    Raw(ON);
    
    while (TRUE) {
	struct menu_common MENU;
	int LINES, COLUMNS;

	menu_get_sizes(page, &LINES, &COLUMNS);   
    
	set_mcommon_from_mbxview(&MENU,mailbox);

	clear_error();

	command = XX(mailbox,command,&screen_mangled, page);  

	switch (command) {

	case 0:
	    return 0;       /* QUIT loop */
	case 256:
	    break;         /* DONE */



#ifdef ALLOW_SUBSHELL
	    case '!' : 
		put_cmd_name(catgets(elm_msg_cat, ElmSet, ElmSystemCall,
				     "System call"), TRUE, &MENU,
			     page);
		(void) subshell(mailbox,page);
		menu_ClearScreen(page);
		build_bottom(&MENU, page);
		break;
#endif

	  case '<' : { 
	      int current = get_current(mailbox);

#ifdef ENABLE_CALENDAR
		put_cmd_name(catgets(elm_msg_cat, ElmSet, ElmScanForCalendarCmd,
				     "Scan messages for calendar entries"), 
			     TRUE, &MENU, page);
		if (give_message_data(mailbox,current-1,
				      NULL,NULL,NULL,
				      /* !!! */
				      NO_mime_parse))
		    /* FIX: mime structure is not currently used */
		    scan_calendar(mailbox);
		else {
		    DPRINT(Debug,3,(&Debug, 
				    "give_message_data [%d] fails",current-1));
		}
#else
		store_msg(catgets(elm_msg_cat, ElmSet, ElmCantScanForCalendar,
				  "Can't scan for calendar entries!"));
#endif
	  }
	      break;

	  case 'b' : {
	      int current = get_current(mailbox);
	      struct header_rec *hdr;
	      FILE              *F;

	      put_cmd_name(catgets(elm_msg_cat, ElmSet, ElmBounceMessage,
				   "Bounce message"), TRUE, &MENU,
			   page);
	      
	      if (give_message_data(mailbox,current-1,
				    &hdr,&F,NULL,
				    /* !!! */
				    NO_mime_parse))
		  remail(hdr,F, aview, page);
	      else {
		  DPRINT(Debug,3,(&Debug, 
				  "give_message_data [%d] fails",current-1));
	      }
	  }
	      break;
	      
	  case 'd' : 
	      delete_msg(TRUE, FALSE, &MENU, *header_page); 
		/* really delete it, silent */
		
		       if (! resolve_mode)
			 store_msg(catgets(elm_msg_cat, ElmSet, ElmMessageMarkedForDeleteion,
				"Message marked for deletion."));
		       else
			 goto next_undel_msg;
		       break;

#ifdef USE_PGP
	  case ctrl('E'): {
	      int current = get_current(mailbox);
	      struct header_rec *hdr;
	      FILE *F;
	      
	      put_cmd_name("Extract PGP public key", TRUE,&MENU,
			   page); /* ??? */
	      
	      if (give_message_data(mailbox,current-1,
				      &hdr,&F,NULL,
				      /* !!! */
				    NO_mime_parse)) { 
		  if (hdr->pgp & PGP_PUBLIC_KEY) {
		      /* FIX: mime structure is not currently used */
		      pgp_extract_public_key (hdr,F, page);

		      if (menu_need_redraw(page))   /* ? ? ? Is this correct */
			  put_border(page);

		  } else
		      store_msg("This message does not contain any keys!");
	      } else {
		  DPRINT(Debug,3,(&Debug, 
				  "give_message_data [%d] fails",current-1));
		  
	      }
	  }
	      break;

	  case ctrl('F'):
	      put_cmd_name(catgets(elm_msg_cat, ElmSet,
				   ElmPassphraseForget,
				   "Forget passphrase"),
			   TRUE, &MENU, page);
	      forget_passphrase();
	    break;
#endif
	    case 'f' : {
		int current = get_current(mailbox);
		struct header_rec *hdr;
		FILE *F;
		
		put_cmd_name(catgets(elm_msg_cat, ElmSet, ElmForwardMessage,
				     "Forward message"), TRUE, &MENU,
			     page);

		forward(current-1,mailbox, aview, page);
			    
		if (menu_need_redraw(page))    /* ? ? ? Is this correct? */
		    put_border(page);
		
	    }
		break;

	    case 'g' : {
		int current = get_current(mailbox);
		struct header_rec *hdr;
		FILE *F;
		
		put_cmd_name(catgets(elm_msg_cat, ElmSet, ElmGroupReply,
				     "Group reply"), TRUE, &MENU,
			     page);

		(void) reply_to_everyone(current-1,mailbox, aview, page);
	    }
	    break;

	  case 'h' : screen_mangled = 0;
	      {
		  int current = get_current(mailbox);

		  struct header_rec *hdr;
		  FILE *F;
		  
		  if (give_message_data(mailbox,current-1,
					  &hdr,&F,NULL,
					  mime_parse_routine)) {
		      if (elm_filter) { 
			  elm_filter = 0; 
			  intbuf = show_msg(hdr,F,current,
					    get_message_count(mailbox),
					    page);
			  elm_filter = 1;
			  return(intbuf);
		      } else
			  return(show_msg(hdr,F,current,
					  get_message_count(mailbox),
					  page)); 
		  }  else {
		      DPRINT(Debug,3,(&Debug, 
				      "give_message_data [%d] fails",current-1));
		  }
	      }
	      break;

	    case 'q' :
	    case 'i' : {
		int current = get_current(mailbox);

		(void) get_page(current, &MENU, header_page);
		clear_error();		/* zero out pending msg   */
		screen_mangled = 0;
		return(0);		/* avoid <return> looping */
	    }
		
	  next_undel_msg :	/* a target for resolve mode actions */

	  case DOWN_MARK :
	  case ' ' :
	  case 'j' :
	  case 'n' : {
		    
	      int current = get_current(mailbox);
	      screen_mangled = 0;
	      if((i=next_message(current-1, TRUE, &MENU)) != -1) {
		  struct header_rec *hdr;
		  FILE *F;
		  
		  current = i+1;
		  set_current(mailbox,current);
		  
		  if (give_message_data(mailbox,current-1,
					  &hdr,&F,NULL,
					  mime_parse_routine))    
		      return(show_msg(hdr,F,current,
				      get_message_count(mailbox),
				      page));
		  else {
		      DPRINT(Debug,3,(&Debug, 
				      "give_message_data [%d] fails",current-1));
		  }
	      }
	      return(0);
	  }

	  case ctrl('N'):  {
		  int current = get_current(mailbox);

		  screen_mangled = 0;
		  for (i = next_message(current-1, TRUE, &MENU);
			    i != -1;
		       i = next_message(i, TRUE, &MENU)) {
		      struct header_rec *hdr = give_header(mailbox,i);

		      if(hdr && ison(hdr->status, UNREAD))
			  break;
		  }
		  if (i != -1) {
		      struct header_rec *hdr;
		      FILE *F;
		      int current = i+1;

		      set_current(mailbox,current);

		      if (give_message_data(mailbox,current-1,
					      &hdr,&F,NULL,
					      mime_parse_routine))
			  return(show_msg(hdr,F,current,
					  get_message_count(mailbox),
					  page));
		      else {
			  DPRINT(Debug,3,(&Debug, 
					  "give_message_data [%d] fails",current-1));
		      }
		  }
		  else return (0);
	      }
next_msg:
	    case 'J' : {
		int current = get_current(mailbox);
		screen_mangled = 0;

		if((i=next_message(current-1, FALSE, &MENU)) != -1) {
		    struct header_rec *hdr;
		    FILE *F;

		    current = i+1;
		    set_current(mailbox,current);
			
		    if (give_message_data(mailbox,current-1,
					    &hdr,&F,NULL,
					    mime_parse_routine))
			return(show_msg(hdr,F,current,
					get_message_count(mailbox),
					page));
		    else {
			DPRINT(Debug,3,(&Debug, 
					"give_message_data [%d] fails",current-1));
		    }
		}
		return(0);
	    }

	  case UP_MARK:
	  case 'k' : {
	      int current = get_current(mailbox);

	      screen_mangled = 0;
	      if((i=prev_message(current-1, TRUE, &MENU)) != -1) {
		  struct header_rec *hdr;
		  FILE *F;
		  
		  current = i+1;
		  set_current(mailbox,current);

		  if (give_message_data(mailbox,current-1,
					  &hdr,&F,NULL,
					  mime_parse_routine))
		      return(show_msg(hdr,F,current,
				      get_message_count(mailbox),
				      page));
		  else {
		      DPRINT(Debug,3,(&Debug, 
				      "give_message_data [%d] fails",current-1));
		  }
	      }
	      return(0);
	  }

	    case 'K' : {
		int current = get_current(mailbox);

		screen_mangled = 0;
		if((i=prev_message(current-1, FALSE, &MENU)) != -1) {
		    struct header_rec *hdr;
		    FILE *F;

		    current = i+1;
		    set_current(mailbox,current);

		    if (give_message_data(mailbox,current-1,
					    &hdr,&F,NULL,
					    mime_parse_routine))
			return(show_msg(hdr,F,current,
					get_message_count(mailbox),
					page));
		    else {
			DPRINT(Debug,3,(&Debug, 
					"give_message_data [%d] fails",current-1));
		    }
		}
		return(0);
	    }

	    case 'm' : {
		int current = get_current(mailbox);
		struct header_rec *hdr;
		FILE *F;

		put_cmd_name(catgets(elm_msg_cat, ElmSet, ElmMailMessage,
				     "Mail message"), TRUE, &MENU,
			     page);
		
		send_msg_l(current-1,
			   NULL,NULL,NULL, MAIL_EDIT_MSG, 
			   allow_forms, 
			   mailbox, aview,
			   page);
    
		if (menu_need_redraw(page)) /* ? ? ? Is this correct? */
		    put_border(page);
	    }
		break;

	    case 'p' : {
		int current = get_current(mailbox);
		
		put_cmd_name(catgets(elm_msg_cat, ElmSet, ElmPrintMessage,
				     "Print message"), FALSE, &MENU,
			     page);
#if DISABLE_PRINT
	    lib_error(FRM("p)rint currently disabled, try P)rint text instead"));
#else
	    if (give_message_data(mailbox,current-1,
				  NULL,NULL,NULL,
				  /* !!! */
				  NO_mime_parse))
		/* FIX: mime structure is not used */
		(void) print_msg(FALSE, mailbox, page);
	    else {
		DPRINT(Debug,3,(&Debug, 
				"give_message_data [%d] fails",current-1));
		
	    }
#endif
	    }
	    break;

	    case 'P' : {
		int current = get_current(mailbox);

		put_cmd_name(catgets(elm_msg_cat, ElmSet, 
					    ElmPrintTextMessage,
				     "Print text of message"), FALSE, &MENU,
			     page);

		if (give_message_data(mailbox,current-1,
					NULL,NULL,NULL,
					mime_parse_routine))
		    (void) print_text(FALSE, mailbox, *header_page, page);
		else {
		    DPRINT(Debug,3,(&Debug, 
				    "give_message_data [%d] fails",current-1));
		    
		}
	    }
	    break;

	    case 'r' : {
		int current = get_current(mailbox);
		struct header_rec *hdr;
		FILE *F;
		
		put_cmd_name(catgets(elm_msg_cat, ElmSet, ElmReplyToMessage,
				     "Reply to message"), TRUE, &MENU,
			     page);

		reply(current-1,mailbox, aview, page);
		if (menu_need_redraw(page))  /* ? ? ? Is this correct? */
		    put_border(page);


	    }
		break;

	    case '>' :
	    case 'C' :
	    case 's' : {
		int current = get_current(mailbox);
		
		put_cmd_name((command != 'C' ? 
			      cap_save_word : cap_copy_word), TRUE, &MENU,
			     page);

		if (give_message_data(mailbox,current-1,
					NULL,NULL,NULL,
				      NO_mime_parse)) {

		    save(TRUE, (command != 'C'),0,
			 mailbox, aview, *header_page,
			 page);
		    
		    if (menu_need_redraw(page))  /* ? ? ? Is this correct? */
			put_border(page);

		} else {
		    DPRINT(Debug,3,(&Debug, 
				    "give_message_data [%d] fails",current-1));
		    
		}

		if (resolve_mode && command != 'C')
		    goto next_undel_msg;
	    }
		break;
	    case 'S' : {
		int current = get_current(mailbox);
		
		put_cmd_name(catgets(elm_msg_cat, ElmSet, ElmSaveText,
				     "Save text"), TRUE, &MENU,
			     page);
		
		if (give_message_data(mailbox,current-1,
					NULL,NULL,NULL,
				      mime_parse_routine)) {
		    save(TRUE, 0,1, mailbox, aview,
			 *header_page,
			 page);

		    if (menu_need_redraw(page))  /* ? ? ? Is this correct? */
			put_border(page);

		} else {
		    DPRINT(Debug,3,(&Debug, 
				    "give_message_data [%d] fails",current-1));
		    
		}
	    }
		break;

	    case 'T' :
	   case 't' : 
	       istagged=tag_message(FALSE, &MENU, *header_page);	
		       if(istagged)
			 store_msg(catgets(elm_msg_cat, ElmSet, ElmMessageTagged,
				"Message tagged."));
		       else 
			 store_msg(catgets(elm_msg_cat, ElmSet, ElmMessageUntagged,
				"Message untagged."));
	               if (command == 'T')
			 goto next_undel_msg;
		       break;

	    case 'u' : 
		undelete_msg(FALSE, &MENU, *header_page); /* undelete it, silently */
		       if (! resolve_mode)
			 store_msg(catgets(elm_msg_cat, ElmSet, ElmMessageUndeleted,
				"Message undeleted."));
		       else {
/******************************************************************************
 ** We're special casing the U)ndelete command here *not* to move to the next
 ** undeleted message ; instead it'll blindly move to the next message in the
 ** list.  See 'elm.c' and the command by "case 'u'" for further information.
 ** The old code was:
			 goto next_undel_msg;
*******************************************************************************/
			 goto next_msg;
		       }
		       break;
	  case 'v' : {
	      struct header_rec *hdr;
	      FILE *F;
	      int current = get_current(mailbox);

	      if (current < 1)
		  error ("There are no messages!");
	      else if (give_message_data(mailbox,current-1,
					 &hdr,&F,NULL,mime_parse_routine)) {
		  if ((hdr->status & MIME_MESSAGE)) {
		      
		      mime_warnings(hdr);
		      
		      attach_menu(F, & (hdr->mime_rec), NULL, 
				  hdr->header_charset, NULL,hdr, aview);
		  }
		  else
		      error ("This is not a MIME message!");
	      } else {
		  DPRINT(Debug,3,(&Debug, 
				  "give_message_data [%d] fails",current-1));
	      }
	      return(0);
	  }
	    case 'X' : put_cmd_name(catgets(elm_msg_cat, ElmSet, ElmQuickExit,
					    "Quick Exit"), TRUE, &MENU,
				    page);

		prompt_s(mcommon_give_item(&MENU,m_Prompt));     /* Command name will be reprinted */
		return 'X';   /* Call caller */


	    case 'x' : put_cmd_name(catgets(elm_msg_cat, ElmSet, ElmExit, "Exit"), TRUE,
				    &MENU,
				    page);

		prompt_s(mcommon_give_item(&MENU,m_Prompt));     /* Command name will be reprinted */
		return 'x';    /* Call caller */

	    
	    case ctrl('J'):
	    case ctrl('M'):  {
		int current = get_current(mailbox);
		struct header_rec *hdr;
		FILE *F;

		screen_mangled = 0;
		if (give_message_data(mailbox,current-1,
					&hdr,&F,NULL,
					mime_parse_routine))
		    return(show_msg(hdr,F,current,
				    get_message_count(mailbox),
				    page));
		else {
		    DPRINT(Debug,3,(&Debug, 
				    "give_message_data [%d] fails",current-1));
		}
		return 0;
	    }

	  case EOF    : 
	      return EOF;   /* FAILURE */
	      

	    default  : putchar((char) 007);	/* BEEP! */
	  }

	  /* display prompt */
	  if (screen_mangled) {
	      /* clear what was left over from previous command 
	       * and display last generated message.
	       */
	      int lin,col;

	      prompt_s(mcommon_give_item(&MENU,m_Prompt));
	      GetXYLocation(&lin,&col);

	      CleartoEOS();
	      PutLineX(LINES-4, 45, 
		        CATGETS(elm_msg_cat, ElmSet, ElmUseIToReturnIndex,
				"(Use 'i' to return to index.)"));


	      /* TODO: Fix this! */
	      Centerline(LINES-1, msg_line, page);
	      MoveCursor(lin, col);
	  } else {
	      /* display bottom line prompt with last generated message */
	      MoveCursor(LINES-1, 0);
	      CleartoEOS();
	      StartBold();
	      Write_to_screen(CATGETS(elm_msg_cat, ElmSet, ElmCommandLine,
				      "%s Command ('i' to return to index): "),
			      msg_line);
	      EndBold();
	  }
	  *msg_line = '\0';	/* null last generated message */

	  command = menu_ReadCh(page,
				'i'|READCH_CURSOR);   /* get next command from user */
	}
}

void put_cmd_name(command, will_mangle, menu, page)
     char *command;
     int will_mangle;
     struct menu_common *menu;
     struct menu_context  *page;
{

	/* If screen is or will be mangled display the command name 
	 * and erase the bottom of the screen.
	 * But first if the border line hasn't yet been drawn, draw it.
	 */
	if(will_mangle && !screen_mangled) {
	    build_bottom(menu,page);
	    screen_mangled = TRUE;
	}
	if(screen_mangled) {
	    prompt_s(mcommon_give_item(menu, m_Prompt));

	    Write_to_screen(FRM("%s"), command);
	    CleartoEOS();
	}
}

void put_border(page)
     struct menu_context  *page;
{
    int LINES, COLUMNS;

    menu_get_sizes(page, &LINES, &COLUMNS);   


	 PutLine0(LINES-5, 0, 
"--------------------------------------------------------------------------\n");
}

void build_bottom(menu, page)
     struct menu_common *menu;
     struct menu_context  *page;
{
    int LINES, COLUMNS;

    menu_get_sizes(page, &LINES, &COLUMNS);   

    MoveCursor(LINES-5, 0);
    CleartoEOS();
    put_border(page);
    prompt_s(mcommon_give_item(menu,m_Prompt));
    PutLineX(LINES-4, 45, 
	      CATGETS(elm_msg_cat, ElmSet, ElmUseIToReturnIndex,
		      "(Use 'i' to return to index.)"));

}


/*
 * Local Variables:
 *  mode:c
 *  c-basic-offset:4
 *  buffer-file-coding-system: iso-8859-1
 * End:
 */
